from ctypes import CFUNCTYPE
import ctypes.util
import platform
from typing import Iterable

import fontfinder

class CTypesLibrary:
    IN:  int    = 1
    OUT: int    = 2
    IN0: int    = 4

    def __init__(self, library_loader, library_name: str, find_library: bool = False, alt_pathname: str = None):
        '''`alt_pathname` is an alternative pathname to try if a library named `library_name` cannot be found.
        '''
        if find_library:
            library_name = ctypes.util.find_library(library_name)
            if library_name is None:
                library_name = alt_pathname
        self.lib = library_loader.LoadLibrary(library_name)

    # Inspired by https://www.cs.unc.edu/~gb/blog/2007/02/11/ctypes-tricks/
    def prototype(self, functype, result_type, func_name, *arg_items):
        '''
        Each arg_item should be
        (in_or_out_const, arg_type[, param_name_str[, default_value]])
        '''
        arg_types = []
        param_flags = []
        for arg_item in arg_items:
            arg_types.append(arg_item[1])
            param_flag = [arg_item[0]]
            if len(arg_item) > 2:
                param_flag.append(arg_item[2])
            if len(arg_item) > 3:
                param_flag.append(arg_item[3])
            param_flags.append(tuple(param_flag))
        return functype(result_type, *arg_types)((func_name, self.lib), tuple(param_flags))

    def c_prototype(self, result_type, func_name, *arg_items):
        '''
        Each arg_item should be
        (in_or_out_const, arg_type[, param_name_str[, default_value]])
        '''
        return self.prototype(CFUNCTYPE, result_type, func_name, *arg_items)

    def w_prototype(self, result_type, func_name, *arg_items):
        '''
        Each arg_item should be
        (in_or_out_const, arg_type[, param_name_str[, default_value]])
        '''
        from ctypes import WINFUNCTYPE
        return self.prototype(WINFUNCTYPE, result_type, func_name, *arg_items)


class FontPlatform:
    def __init__(self):
        pass

    def all_installed_families():
        '''Return a list of font family names available on this platform.'''
        pass

    def install_fonts(self, font_infos: Iterable[fontfinder.FontInfo]) -> None:
        '''Install the font files in `font_infos`. The `downloaded_path` of each `FontInfo` must point to the
        actual font file in the filesystem.
        
        Font are installed to the user font collection, rather than the system-wide font collection.'''
        pass

    def uninstall_fonts(self, font_infos: Iterable[fontfinder.FontInfo]) -> None:
        '''Uninstall the font files in `font_infos`. The fonts must exist in the user font collection, otherwise they
        will not be uninstalled.
        '''
        pass

    def known_platform_fonts(self) -> list[fontfinder.FontInfo]:
        '''Returns a list of FontInfo objects for all fonts known to this library that are unique to this platform.
        '''
        pass

    def set_platform_prefs(self, font_finder: 'fontfinder.FontFinder') -> None:
        pass


def get_font_platform():
    if platform.system() == "Darwin":
        import fontfinder._platforms.mac
        return fontfinder._platforms.mac.MacPlatform()
    elif platform.system() == "Windows":
        import fontfinder._platforms.windows
        return fontfinder._platforms.windows.WindowsPlatform()
    else:
        raise fontfinder.UnsupportedPlatformException()




# Code in this module derived from FindSystemFontFilename project, under the MIT Licence:
# https://github.com/moi15moi/FindSystemFontsFilename/
#
# See ACKNOWLEDGEMENTS file.
#
import platform
if platform.system() == "Darwin":
    from ctypes import c_bool, c_char_p, c_long, c_uint32, c_void_p, create_string_buffer
    import ctypes.util
    from pathlib import Path
    import os
    import shutil

    from packaging import version

    from fontfinder import FontFinder, FontFinderException
    import fontfinder._platforms
    import fontfinder.filters


    USER_FONT_DIR = Path("~/Library/Fonts").expanduser()


    class MacPlatform(fontfinder._platforms.FontPlatform):
        def all_installed_families(self):        
            if version.Version(platform.mac_ver()[0]) < version.Version('10.6'):
                raise Exception("fontfinder.mac.all_installed_families() only supported by macOS 10.6 or later")
            
            cf = CoreFoundationLibrary()
            ct = CoreTextLibrary()

            font_collection = ct.CTFontCollectionCreateFromAvailableFonts(None)
            font_array = ct.CTFontCollectionCreateMatchingFontDescriptors(font_collection)
            font_array_len = cf.CFArrayGetCount(font_array)
            family_names = set()
            for i in range(font_array_len):
                font_descriptor = cf.CFArrayGetValueAtIndex(font_array, i)

                family_cfstr = ct.CTFontDescriptorCopyAttribute(font_descriptor, ct.kCTFontFamilyNameAttribute)
                family_name = cf.cf_string_ref_to_python_str(family_cfstr)
                cf.CFRelease(family_cfstr)

                # style_cfstr = ct.CTFontDescriptorCopyAttribute(font_descriptor, ct.kCTFontStyleNameAttribute)
                # style_name = cf.cf_string_ref_to_python_str(style_cfstr)
                # cf.CFRelease(style_cfstr)

                # display_cfstr = ct.CTFontDescriptorCopyAttribute(font_descriptor, ct.kCTFontDisplayNameAttribute)
                # display_name = cf.cf_string_ref_to_python_str(display_cfstr)
                # cf.CFRelease(display_cfstr)

                # postscript_cfstr = ct.CTFontDescriptorCopyAttribute(font_descriptor, ct.kCTFontNameAttribute)
                # postscript_name = cf.cf_string_ref_to_python_str(postscript_cfstr)
                # cf.CFRelease(postscript_cfstr)

                family_names.add(family_name)
            cf.CFRelease(font_array)
            cf.CFRelease(font_collection)
            return sorted(list(family_names))

        def install_fonts(self, font_infos):
            for font_info in font_infos:
                if font_info.downloaded_path is not None and font_info.downloaded_path != Path():
                    shutil.copy2(font_info.downloaded_path, USER_FONT_DIR)
                else:
                    raise FontFinderException("Can't install font without a path to the downloaded font file")

        def uninstall_fonts(self, font_infos):
            for font_info in font_infos:
                if font_info.filename is None:
                    raise FontFinderException("Can't uninstall font without a filename")
                os.remove(USER_FONT_DIR / font_info.filename)

        def known_platform_fonts(self) -> list[fontfinder.FontInfo]:
            # Use Apple Color Emoji for emoji.
            font_infos = []
            font_infos.append(fontfinder.FontInfo(main_script="Common", script_variant="Emoji",
                            family_name="Apple Color Emoji",
                            subfamily_name="Regular", postscript_name="AppleColorEmoji",
                            form=fontfinder.FontForm.UNSET, width=fontfinder.FontWidth.NORMAL,
                            weight=fontfinder.FontWeight.REGULAR, style=fontfinder.FontStyle.UPRIGHT,
                            format=fontfinder.FontFormat.TTF, build=fontfinder.FontBuild.UNSET,
                            url=""))
            return font_infos
        
        def set_platform_prefs(self, font_finder: FontFinder) -> None:
            # Use Apple Color Emoji for emoji.
            font_finder.family_prefs[("Common", "Emoji")] = [fontfinder.filters.attr_in("family_name",
                                                                                        ["Apple Color Emoji"])]


    class CoreFoundationLibrary(fontfinder._platforms.CTypesLibrary):
        def __init__(self):
            super().__init__(ctypes.cdll, "CoreFoundation", True,
                            "/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")
            # Note hack for compatibility with macOS 11.0 and later
            # From: https://github.com/pyglet/pyglet/blob/a44e83a265e7df8ece793de865bcf3690f66adbd/pyglet/libs/darwin/cocoapy/cocoalibs.py#L10-L14

            self.CFRelease = self.c_prototype(
                None, "CFRelease", (self.IN, c_void_p, "cf_object"))

            self.CFArrayGetCount = self.c_prototype(
                c_long, "CFArrayGetCount", (self.IN, c_void_p, "cf_array"))

            self.CFArrayGetValueAtIndex = self.c_prototype(
                c_void_p, "CFArrayGetValueAtIndex", (self.IN, c_void_p, "cf_array"), (self.IN, c_long, "index"))

            self.CFStringGetMaximumSizeForEncoding = self.c_prototype(
                c_long, "CFStringGetMaximumSizeForEncoding", (self.IN, c_long, "length"), (self.IN, c_uint32, "encoding"))

            self.CFStringGetLength = self.c_prototype(
                c_long, "CFStringGetLength", (self.IN, c_void_p, "the_string"))

            self.CFStringGetCString = self.c_prototype(
                c_bool, "CFStringGetCString", (self.IN, c_void_p, "the_string"),
                                            (self.IN, c_char_p, "buffer"),
                                            (self.IN, c_long,   "buffer_size"),
                                            (self.IN, c_uint32, "encoding")
            )

            self.kCFStringEncodingUTF8 = c_uint32(0x08000100)

        def cf_string_ref_to_python_str(self, cf_string_ref: c_void_p):
            cf_str_len = self.CFStringGetLength(cf_string_ref)
            buffer_size = self.CFStringGetMaximumSizeForEncoding(cf_str_len, self.kCFStringEncodingUTF8)
            buffer = create_string_buffer(buffer_size)
            success = self.CFStringGetCString(cf_string_ref, buffer, buffer_size, self.kCFStringEncodingUTF8)
            if not success:
                raise Exception("Couldn't encode string as UTF-8 into buffer")
            python_str = buffer.raw.strip(b'\x00').decode(encoding='utf-8')
            return python_str


    class CoreTextLibrary(fontfinder._platforms.CTypesLibrary):
        def __init__(self):
            super().__init__(ctypes.cdll, "CoreText", True,
                            "/System/Library/Frameworks/CoreText.framework/CoreText")
            # Note hack for compatibility with macOS greater or equals to 11.0.
            # From: https://github.com/pyglet/pyglet/blob/a44e83a265e7df8ece793de865bcf3690f66adbd/pyglet/libs/darwin/cocoapy/cocoalibs.py#L520-L524

            self.CTFontCollectionCreateFromAvailableFonts = self.c_prototype(
                c_void_p, "CTFontCollectionCreateFromAvailableFonts", (self.IN, c_void_p, "options"))

            self.CTFontCollectionCreateMatchingFontDescriptors = self.c_prototype(
                c_void_p, "CTFontCollectionCreateMatchingFontDescriptors", (self.IN, c_void_p, "font_collection"))
            
            self.CTFontDescriptorCopyAttribute = self.c_prototype(
                c_void_p, "CTFontDescriptorCopyAttribute", (self.IN, c_void_p, "font_descriptor"),
                                                        (self.IN, c_void_p, "attribute_name")
            )

            self.kCTFontFamilyNameAttribute = c_void_p.in_dll(self.lib, "kCTFontFamilyNameAttribute")
            self.kCTFontStyleNameAttribute = c_void_p.in_dll(self.lib, "kCTFontStyleNameAttribute")
            self.kCTFontDisplayNameAttribute = c_void_p.in_dll(self.lib, "kCTFontDisplayNameAttribute")
            self.kCTFontNameAttribute = c_void_p.in_dll(self.lib, "kCTFontNameAttribute")



'''
Useful filter factories for expressing font preferences for `FontFinder.family_prefs` and
`FontFinder.family_font_prefs`.
'''

def attr_in(attr_name, collection):
    '''A filter factory. Returns a filter function that takes a single argument `obj` and returns True
    if `obj.attr_name` is in `collection`, else False.
    '''
    def filter(obj):
        return getattr(obj, attr_name) in collection
    return filter

def attr_not_in(attr_name, collection):
    '''A filter factory. Returns a filter function that takes a single argument `obj` and returns True
    if `obj.attr_name` is not in `collection`, else False.
    '''
    def filter(obj):
        return getattr(obj, attr_name) not in collection
    return filter

def attr_contains(attr_name, collection):
    '''A filter factory. Returns a filter function that takes a single argument `obj` and returns True
    if any of the items in `collection` are in `obj.attr_name`. Otherwise returns False.
    '''
    def filter(obj):
        return any(map(lambda item: item in getattr(obj, attr_name), collection))
    return filter

def attr_not_contains(attr_name, collection):
    '''A filter factory. Returns a filter function that takes a single argument `obj` and returns True
    if none of the items in `collection` are in `obj.attr_name`. Otherwise returns False.
    '''
    def filter(obj):
        return not any(map(lambda item: item in getattr(obj, attr_name), collection))
    return filter

def attr_contains_str(attr_name, str_collection):
    '''A filter factory. Returns a filter function that takes a single argument `obj` and returns True
    if any of the strings in `str_collection` are in the string conversion of `obj.attr_name`, when all strings are
    casefolded. Otherwise returns False.
    '''
    def filter(obj):
        return any(map(lambda s: s.casefold() in str(getattr(obj, attr_name)).casefold(), str_collection))
    return filter

def attr_not_contains_str(attr_name, str_collection):
    '''A filter factory. Returns a filter function that takes a single argument `obj` and returns True
    if none of the strings in `str_collection` are in the string conversion of `obj.attr_name`, when all strings are
    casefolded. Otherwise returns False.
    '''
    def filter(obj):
        return not any(map(lambda s: s.casefold() in str(getattr(obj, attr_name).casefold()), str_collection))
    return filter



'''
## Overview

**fontfinder is a Python package for finding and installing fonts for Unicode scripts. It's useful
when generating documents that must specify a font family and will be viewed across multiple platforms.**
For now, `fontfinder` mostly locates fonts in the [Google Noto font collection](https://fonts.google.com/noto).

Font enumeration and installation is currently supported on macOS (using CoreText) and Windows (using DirectWrite).

Most functionality is provided by instantiating the `FontFinder` class.

## Examples

```python
>>> from fontfinder import FontFinder
>>> ff = FontFinder()
>>> text = 'الشمس (رمزها: ☉) هي النجم المركزي للمجموعة الشمسية.' # From Arabic Wikipedia article about the Sun
>>> ff.analyse(text)
TextInfo(main_script='Arabic', script_variant='', emoji_count=1, script_count=Counter({'Arabic': 39, 'Common': 12}))
>>> known_families = ff.find_families(text) # Available font families for the given text
>>> print(known_families)
['Noto Kufi Arabic', 'Noto Naskh Arabic', 'Noto Naskh Arabic UI', 'Noto Sans Arabic']
>>> preferred_family = ff.find_family(text) # Selects a single preferred font family for the text
>>> print(preferred_family)
Noto Naskh Arabic
>>> print(ff.installed_families(preferred_family)) # The font family is not yet installed
[]
>>> family_fonts = ff.find_family_fonts(preferred_family)
>>> print([font_info.postscript_name for font_info in family_fonts]) # The individual fonts in the family
['NotoNaskhArabic-Bold', 'NotoNaskhArabic-Medium', 'NotoNaskhArabic-Regular', 'NotoNaskhArabic-SemiBold']
>>> from tempfile import TemporaryDirectory
>>> tempdir = TemporaryDirectory()
>>> fonts_for_download = ff.find_family_fonts_to_download(preferred_family) # The indiv fonts that can be downloaded
>>> fonts_for_install = ff.download_fonts(fonts_for_download, tempdir.name) # The downloaded fonts
>>> ff.install_fonts(fonts_for_install) # Actually install each individual font
>>> print(ff.installed_families(preferred_family)) # The font family is now installed
['Noto Naskh Arabic']
>>> ff.uninstall_fonts(fonts_for_install) # Uninstall each individual font
>>> print(ff.installed_families(preferred_family)) # The font family is once again not installed
[]
>>> tempdir.cleanup()
```

## Attribution

Parts of this library are derived from code in the FindSystemFontsFilename
library by moi15moi (https://github.com/moi15moi/FindSystemFontsFilename/), under the MIT Licence.

See ACKNOWLEDGEMENTS for more detail.

## Installation

   `pip install fontfinder`

## Build Instructions

Use these instructions if you’re building from the source. `fontfinder` has been developed on Python 3.10, but should
work on other versions as well.

1. `git clone https://github.com/multiscript/fontfinder/`
1. `cd fontfinder`
1. `python3 -m venv venv` (Create a virtual environment.)
   - On Windows: `python -m venv venv`
1. `source venv/bin/activate` (Activate the virtual environment.)
   - In Windows cmd.exe: `venv\Scripts\\activate.bat`
   - In Windows powershell: `.\\venv\Scripts\Activate.ps1` You may first need to run `Set-ExecutionPolicy -ExecutionPolicy RemoteSigned -Scope CurrentUser`
1. For development work...
   - `pip install -e .` (Creates an editable local install)
1. ...or to build the package:
   - `pip install build`
   - `python -m build`

## Top-Level Objects
'''
from collections import Counter
import json
from pathlib import Path
import tempfile
from typing import Iterable

import platformdirs
import requests
import unicodedataplus as udp

from fontfinder.model import TextInfo
from fontfinder.filters import *
from fontfinder.model import *
from fontfinder import _platforms


_REF_DATA_DIR_PATH = Path(__file__, "../data").resolve()
'''Path to reference font data (within package)'''
_REF_DATA_DIR_PATH.mkdir(parents=True, exist_ok=True)

_USER_DATA_DIR_PATH = platformdirs.user_data_path("fontfinder")
'''Path to user data path for fontfinder (outside of package)'''
_USER_DATA_DIR_PATH.mkdir(parents=True, exist_ok=True)

_SMALL_UNIHAN_PATH = Path(_REF_DATA_DIR_PATH, "small_unihan.json").resolve()
'''Path to subset of Unihan data needed for CJK font selection.'''

_SCRIPT_METADATA_URL = "https://raw.githubusercontent.com/unicode-org/cldr-json/main/cldr-json/cldr-core/scriptMetadata.json"
'''URL of script metadata from the Unicode Common Locale Data Repository (CLDR).'''

_SCRIPT_METADATA_PATH = Path(_REF_DATA_DIR_PATH, "scriptMetadata.json").resolve()
'''Path to local copy of script metadata from the Unicode Common Locale Data Repository (CLDR).'''

# We wait until now to import Noto data so that data path constants above are set.
from fontfinder import noto 


class FontFinder:
    '''Main class for accessing this library's functionality.'''
    def __init__(self):
        self._all_known_fonts = None
        self._small_unihan_data_private = None
        self._script_metadata_private = None
        
        self.max_analyse_chars: int = 2048
        '''Maximum number of characters examined by `FontFinder.analyse().'''
        
        self.zh_hant_use_hk = False
        '''If True, `FontFinder.analyse()` selects Hong Kong rather than Taiwanese fonts for Traditional
        Chinese script.'''

        self.family_prefs = {}
        '''Font preferences for selecting a single font-family for some text. This attribute is a dictionary of lists
        of filter functions. See `set_prefs()` for more info.'''

        self.family_font_prefs = {}
        '''Font preferences for selecting font files within a given font-family. This attribute is a dictionary of
        lists of filter functions. See `set_prefs()` for more info.'''

        self.set_prefs()
        self.set_platform_prefs()

    def set_prefs(self) -> None:
        '''Sets the font preferences. See the source code for this method to examine the built-in preferences
        that `fontfinder` uses 'out-of-the-box'. These can be replaced either by overriding this method, or editing
        the `font_family_prefs` and `family_member_prefs` instance attributes.
        
        Font preferences are dictionaries of lists of filter functions. The dictionary keys are one of:
        - A tuple of `(main_script, script_variant)`. Preferences under these keys will only apply to that particular
          script and variant combination.
        - the `fontfinder.ANY_SCRIPT` object. Preferences under this key will apply to any script.
        
        Preferences for particular script/variant combinations are applied before preferences for `ANY_SCRIPT`.
        
        The dictionary values are lists of filter functions. Filter functions take a single
        `fontfinder.model.FontInfo` argument and return True if the argument should be included in the filtered list.
        It's usually easiest to create filter functions using the filter factores in the `fontfinder.filters` module.
        
        If applying a preference would result in all remaining fonts being excluded, the preference is ignored.

        Some example preferences:
        ```python
        # For Arabic, prefer the more traditional Naskh form
        self.family_prefs[("Arabic", "")] = [attr_in("family_name", ["Noto Naskh Arabic"])]

        # Prefer sans-serif fonts, and exclude mono, display and UI forms where possible.
        self.family_prefs[ANY_SCRIPT] = [attr_in("form",           [FontForm.SANS_SERIF]),
                                         attr_not_contains("tags", [FontTag.MONO, FontTag.DISPLAY, FontTag.UI])]
        ```
        '''
        # For Adlam, prefer joined to unjoined.
        self.family_prefs[("Adlam", "")] = [attr_in("family_name", ["Noto Sans Adlam"])]
        # For Arabic, prefer the more traditional Naskh form
        self.family_prefs[("Arabic", "")] = [attr_in("family_name", ["Noto Naskh Arabic"])]
        # For Hebrew, prefer the more traditional Serif form
        self.family_prefs[("Hebrew", "")] = [attr_in("family_name", ["Noto Serif Hebrew"])]
        # For Khitan Small Script, prefer Noto Serif Khitan Small Script, as the purpose of the other fonts isn't clear
        self.family_prefs[("Khitan_Small_Script", "")] = [attr_in("family_name",
                                                                 ["Noto Serif Khitan Small Script"])]
        # For Lao, prefer more traditional looped fonts
        self.family_prefs[("Lao", "")] = [attr_contains_str("family_name", ["Looped"])]
        # For Nko, prefer Noto Sans NKo to unjoined
        self.family_prefs[("Nko", "")] = [attr_in("family_name", ["Noto Sans NKo"])]
        # For Nushu, prefer Noto Sans Nushu as it is better for smaller font sizes
        self.family_prefs[("Nushu", "")] = [attr_in("family_name", ["Noto Sans Nushu"])]
        # For Tamil, don't use the Supplement font
        self.family_prefs[("Tamil", "")] = [attr_not_contains_str("family_name", ["Supplement"])]
        # For Thai, prefer more traditional looped fonts, and Noto Sans Thai Looped in particular
        self.family_prefs[("Thai", "")] = [attr_in("family_name", ["Noto Sans Thai Looped"])]
        # Prefer sans-serif fonts, and exclude mono, display and UI forms where possible.
        self.family_prefs[ANY_SCRIPT] = [attr_in("form",           [FontForm.SANS_SERIF]),
                                         attr_not_contains("tags", [FontTag.MONO, FontTag.DISPLAY, FontTag.UI])]
        
        # With a font family, avoid variable fonts, and mono, display and UI fonts. Prefer full builds, and OTF files.
        self.family_font_prefs[ANY_SCRIPT] = [attr_not_in("width",      [FontWidth.VARIABLE]),
                                              attr_not_in("weight",     [FontWidth.VARIABLE]),
                                              attr_not_contains("tags", [FontTag.MONO, FontTag.DISPLAY, FontTag.UI]),
                                              attr_in("build",          [FontBuild.FULL]),
                                              attr_in("build",          [FontBuild.HINTED]),
                                              attr_in("format",         [FontFormat.OTF]),
                                              attr_in("format",         [FontFormat.TTF]),
                                              attr_in("format",         [FontFormat.OTC]),
                                             ]

    def set_platform_prefs(self) -> None:
        '''Sets any platform-specific font preferences. See the source code in the various sub-modules in
        `_platforms` to examine the built-in platform-specific preferences.
        
        This method is called only after `set_prefs()` has set the platform-independant font preferences.'''
        font_platform = _platforms.get_font_platform()
        font_platform.set_platform_prefs(self)       

    def analyse(self, text: str) -> TextInfo:
        '''Analyse an initial portion of `text` for the Unicode scripts it uses. Returns a
        `fontfinder.model.TextInfo` object with the results.

        The number of characters analysed is set by the instance attribute `max_analyse_chars`.

        The attributes of the `TextInfo` result object are set as follows:
        - `main_script`:    Name of the most-frequently-used Unicode script in `text`. This is the [long Unicode
                            script value (known as a property value
                            alias)](https://unicode.org/reports/tr24/#Script_Value_Aliases), rather than the shorter
                            script code.

        - `script_variant`: A secondary string used when the value of `main_script` is insufficient for choosing
                            an appropriate font. This is not a Unicode property, but a scheme only used by
                            `fontfinder`.

        - `emoji_count`:    Count of characters who have either the Emoji Presentation property or the
                            Extended_Pictographic property set (independent of script).

        - `script_count`:   A [collections.Counter](https://docs.python.org/3/library/collections.html#collections.Counter)
                            of the count of each Unicode script in the text. The keys are the string names of each
                            script that appears in the text (including `Common`, `Inherited` and `Unknown`).

        In calculating `main_script`, the script values `Common`, `Inherited`, and `Unknown` are
        ignored. However if `emoji_count` is larger than the rest of the script counts, then `main_script` is set to
        `Common` and `script_variant` is set to `Emoji`. (Most emoji characters have a Unicode script value of
        `Common`.)
        
        If `main_script` is `Han`, some basic language detection is performed, and the `script_variant` is set to
        one of the following language tags:
        - For Simplified Chinese:  `zh-Hans`                   
        - For Traditional Chinese: `zh-Hant` (or `zh-Hant-HK` if the instance attribute `zh_hant_use_hk` is True)
        - For Japanese:            `ja`
        - For Korean:              `ko`
        '''
        # Do the counting
        script_count = Counter()
        unihan_counter = Counter()
        emoji_count = 0
        for char in text[0:min(len(text), self.max_analyse_chars)]:
            script_count[udp.script(char)] += 1
            if udp.is_emoji_presentation(char) or udp.is_extended_pictographic(char):
                emoji_count += 1
            if char in self._small_unihan_data:
                for key in self._small_unihan_data[char].keys():
                    unihan_counter[key] += 1
        
        # Determine main_script and script_variant
        non_generic_count = script_count.copy()
        generic_scripts = ['Common', 'Inherited', 'Unknown']
        for generic_script in generic_scripts:
            del non_generic_count[generic_script]

        if len(non_generic_count) > 0:
            main_script = non_generic_count.most_common(1)[0][0]
            script_variant = ""
        else:
            main_script = ""
            script_variant = ""

        # Handle emoji
        if (len(non_generic_count) == 0 and emoji_count > 0) or \
           (len(non_generic_count) > 0  and emoji_count > non_generic_count.most_common(1)[0][1]):
            main_script = "Common"
            script_variant = "Emoji"

        # Handle Han script
        if main_script == 'Han':
            # Han script can be used by Chinese, Japanese and Korean texts
            if 'Hangul' in script_count:
                # If Hangul characters are present, assume it's Korean
                script_variant = 'ko'
            elif 'Hiragana' in script_count or 'Katakana' in script_count:
                # If Hirogana or Katakana characters are present, assume it's Japanese
                script_variant = 'ja'
            elif unihan_counter['kSimplifiedVariant'] > unihan_counter['kTraditionalVariant']:
                # Traditional Chinese characters have simplified variants, and vice versa.
                # So if there are more simplified variants than traditional, we likely have traditional text,
                # and vice-versa.
                script_variant = 'zh-Hant-HK' if self.zh_hant_use_hk else 'zh-Hant'
            else:
                script_variant = 'zh-Hans'

        return TextInfo(main_script=main_script, script_variant=script_variant, emoji_count=emoji_count,
                        script_count=script_count)

    def find_families(self, str_or_text_info: str | TextInfo) -> list[str]:
        '''Returns a list of the family names (strings) of all fonts known to the library that are suitable for
        `str_or_text_info`. No font family preferences are applied. If no suitable family names are found,
        an empty list is returned.
        
        `str_or_text_info` should either be the text string itself, or a `TextInfo` object returned by `analyse()`.
        '''
        font_infos = self._text_info_to_font_infos(str_or_text_info)
        # We use a dictionary as a set that preserves insertion order, to return families in their original order.
        family_names = {font_info.family_name: 1 for font_info in font_infos}
        return list(family_names.keys())

    def find_family(self, str_or_text_info: str | TextInfo) -> str | None:
        '''Returns the family name (a string) of the preferred font family for `str_or_text_info`.
        The preferred font family is determined applying the filter functions in `family_prefs`.
        If, after applying these filters, more than one family remains, the first family is returned.
        If no preferred font family is found, None is returned.
        
        `str_or_text_info` should either be the text string itself, or a `TextInfo` object returned by `analyse()`.
        '''
        font_infos = self._text_info_to_font_infos(str_or_text_info)
        if len(font_infos) == 0:
            return None
        count_func = lambda font_infos: len({font_info.family_name for font_info in font_infos})
        font_infos = self._apply_pref_dict(font_infos[0].main_script, font_infos[0].script_variant,
                                           self.family_prefs, count_func, font_infos)
        family_name = font_infos[0].family_name
        return family_name

    def find_family_fonts(self, family_name_or_names: str | Iterable[str],
                          main_script:str = None, script_variant:str = None) -> list[FontInfo]:
        '''Returns a list of `FontInfo` objects for font files matching the given font family names.
        The list is filtered using the functions in `family_font_prefs`.
         
        `family_name_or_names` can be a single font family name, or an iterable of family names.

        Some font families match several Unicode scripts. In these cases, `main_script` and `script_variant`
        can optionally be specified, to ensure these fields have the correct value in the returned `FontInfo` list.
        Otherwise, `main_script` and `script_variant` will have the first values found within the given font families.
        '''
        family_names = family_name_or_names
        if isinstance(family_names, str):
            family_names = [family_names]
        
        result_font_infos = []
        for family_name in family_names:
            font_infos = self.known_fonts(lambda font_info: font_info.family_name == family_name)
            if len(font_infos) == 0:
                continue
            # font_infos can be duplicated under multiple script variants. If no script and variant is specified, we
            # just pick the first.
            if main_script is None:
                family_main_script = font_infos[0].main_script
            else:
                family_main_script = main_script
            if script_variant is None:
                family_script_variant = font_infos[0].script_variant
            else:
                family_script_variant = script_variant
            font_infos = [font_info for font_info in font_infos if font_info.main_script == family_main_script and \
                                                                font_info.script_variant == family_script_variant]
            count_func = len
            font_infos = self._apply_pref_dict(main_script, script_variant, self.family_font_prefs, count_func,
                                            font_infos)
            result_font_infos.extend(font_infos)
        return result_font_infos

    def find_family_fonts_to_download(self, family_name_or_names: str | Iterable[str],
                                      main_script:str = None, script_variant:str = None) -> list[FontInfo]:
        '''Returns a list of `FontInfo` objects for font files that need to be downloaded. This list is formed
        by finding fonts matching the given family names, where those families are not currently installed,
        the filters in `family_font_prefs` have been applied, and the fonts have download URLs provided.

        If the returned list is empty, then no fonts need to be downloaded, either because the font families are
        already installed, or no download URLs are provided.
        
        `family_name_or_names` can be a single font family name, or an iterable of family names.

        Some font families match several Unicode scripts. In these cases, `main_script` and `script_variant`
        can optionally be specified, to ensure these fields have the correct value in the resulting `FontInfo` list.
        Otherwise, `main_script` and `script_variant` will have the first values found within the given font families.
        '''
        family_names = self.not_installed_families(family_name_or_names)
        font_infos = self.find_family_fonts(family_names, main_script, script_variant)
        font_infos = self.downloadable_fonts(font_infos)
        return font_infos

    def download_fonts(self, font_infos: Iterable[FontInfo], download_dir: str | Path) -> list[FontInfo]:
        '''Downloads the font files in `font_infos`, in preparation for installation. The font files are downloaded
        to `download_dir`. Returns a list of copied `FontInfo` objects where the `downloaded_path` attribute points
        to each new file.'''
        download_dir = Path(download_dir)
        font_infos = self.downloadable_fonts([font_info.copy() for font_info in font_infos])
        for font_info in font_infos:
            response = requests.get(font_info.url, stream=True)
            font_info.downloaded_path = download_dir / font_info.filename
            with open(font_info.downloaded_path, 'wb') as file:
                for chunk in response.iter_content(chunk_size=128):
                    file.write(chunk)
        return font_infos

    def install_fonts(self, font_infos: Iterable[FontInfo]) -> None:
        '''Install the font files in `font_infos`. The `downloaded_path` of each `FontInfo` must point to the
        actual font file in the filesystem.
        
        Font are installed to the user font collection, rather than the system-wide font collection.'''
        font_platform = _platforms.get_font_platform()
        font_platform.install_fonts(font_infos)        
     
    def uninstall_fonts(self, font_infos: Iterable[FontInfo]) -> None:
        '''Uninstall the font files in `font_infos`. The fonts must exist in the user font collection, otherwise they
        will not be uninstalled.
        '''
        font_platform = _platforms.get_font_platform()
        font_platform.uninstall_fonts(font_infos)        

    def is_rtl(self, script_or_text_info: str | TextInfo) -> bool:
        '''Returns True if the text direction of the given Unicode script is right-to-left, otherwise False.
        If the script name is not found, returns False.
        
        `script_or_text_info` must be either the long script name (i.e. the Unicode property value alias), or a
        `TextInfo` object, in which case the `main_script` attribute of the `TextInfo` will be used.
        '''
        long_script = script_or_text_info
        if isinstance(long_script, TextInfo):
            long_script = long_script.main_script

        short_script = udp.property_value_aliases["script"].get(long_script, [None])[0]
        if short_script is None:
            return False
        is_rtl = (self._script_metadata[short_script]["rtl"] == "YES")
        return is_rtl

    def known_fonts(self, filter_func = None) -> list[FontInfo]:
        '''Returns a list of FontInfo objects for all fonts known to this library.
        
        This is a large list, which is cached in memory the first time the method is called.'''
        # Even though noto.get_noto_fonts() can filter on the fly, for now we choose to optimise for speed, rather
        # than memory, by caching the full list of font_infos in memory.
        if self._all_known_fonts is None:
            self._all_known_fonts = []
            self._all_known_fonts.extend(noto.get_noto_fonts())
            font_platform = _platforms.get_font_platform()
            self._all_known_fonts.extend(font_platform.known_platform_fonts())

        return [font_info for font_info in self._all_known_fonts if (filter_func is None or filter_func(font_info))]

    def known_scripts(self, filter_func = None) -> list[str]:
        '''Returns a list of the `main_script` values for all the fonts known to this library.'''
        return sorted(set([info.main_script for info in self.known_fonts(filter_func)]))

    def known_script_variants(self, filter_func = None) -> list[(str, str)]:
        '''Returns a list of `(main_script, script_variant)` tuples for all the fonts known to this library.'''
        # Use a dictionary as an ordered set
        return list({(info.main_script, info.script_variant): 1 for info in self.known_fonts(filter_func)}.keys())

    def all_unicode_scripts(self) -> list[str]:
        '''Returns a list of all script values (property value aliases) in the Unicode standard.'''
        return list(udp.property_value_aliases['script'].keys())

    def scripts_not_known(self) -> list[str]:
        '''Returns a list of all the Unicode script values not supported by the fonts known to this library.'''
        return sorted(set(self.all_unicode_scripts()) - set(self.known_scripts()) -
                      set(["Common", "Inherited", "Unknown"]))

    def all_installed_families(self) -> list[str]:
        '''Returns a list of the family names of all fonts currently installed on the system.
        '''
        font_platform = _platforms.get_font_platform()
        return font_platform.all_installed_families()        

    def installed_families(self, family_name_or_names: str | Iterable[str]) -> list[str]:
        '''For a given font family name or iterable of names, return a filtered list containing just those
        families that are currently installed on the system.'''
        family_names = family_name_or_names
        if isinstance(family_names, str):
            family_names = [family_names]
        all_installed_families = set(self.all_installed_families())
        return [family_name for family_name in family_names if family_name in all_installed_families]

    def not_installed_families(self, family_name_or_names: str | Iterable[str]) -> list[str]:
        '''For a given font family name or iterable of names, return a filtered list containing just those
        families that are not currently installed on the system.'''
        family_names = family_name_or_names
        if isinstance(family_names, str):
            family_names = [family_names]
        all_installed_families = set(self.all_installed_families())
        return [family_name for family_name in family_names if family_name not in all_installed_families]

    def downloadable_fonts(self, font_infos: Iterable[FontInfo]) -> list[FontInfo]:
        '''For a given iterable of `font_infos`, return a filtered list containing just those `FontInfo`s that
        have download URLs provided.'''
        return [font_info for font_info in font_infos if font_info.url is not None and font_info.url != ""]

    @property
    def _small_unihan_data(self):
        if self._small_unihan_data_private is None:
            with open(_SMALL_UNIHAN_PATH, "r", encoding="utf-8") as small_unihan_file:
                self._small_unihan_data_private = json.load(small_unihan_file)
        return self._small_unihan_data_private

    @property
    def _script_metadata(self):
        if self._script_metadata_private is None:
            with open(_SCRIPT_METADATA_PATH, "r", encoding="utf-8") as script_metadata_file:
                self._script_metadata_private = json.load(script_metadata_file)["scriptMetadata"]
        return self._script_metadata_private

    def _text_info_to_font_infos(self, str_or_text_info):
        if isinstance(str_or_text_info, str):
            text_info = self.analyse(str_or_text_info)
        else:
            text_info = str_or_text_info
        font_infos = self.known_fonts(lambda font_info: font_info.main_script == text_info.main_script and \
                                                        font_info.script_variant == text_info.script_variant)
        return font_infos

    def _apply_pref_dict(self, main_script, script_variant, pref_dict, count_func, font_infos):
        # Preferences for particular scripts are applied before preferences for any script
        pref_keys = [(main_script, script_variant), ANY_SCRIPT]
        for pref_key in pref_keys:
            if pref_key in pref_dict:
                font_infos = self._apply_pref_filters(pref_dict[pref_key], count_func, font_infos)
        return font_infos

    def _apply_pref_filters(self, filter_funcs, count_func, font_infos):
        cur_list = font_infos
        count = count_func(cur_list)
        # print(f"Initial ({count})")
        # print([info.url for info in font_infos])
        # print()
        if count < 2 or len(filter_funcs) == 0:
            # We actually don't need to filter.
            return cur_list

        # print("After each filter func")
        for filter_func in filter_funcs:
            new_list = [font_info for font_info in cur_list if filter_func(font_info)]
            count = count_func(new_list)
            # print(count)
            # print([info.url for info in new_list])
            # print()
            if count == 0:
                # This preference was too restrictive, so we ignore it by not updating cur_list
                pass
            elif count == 1:
                # Perfect! Stop filtering
                cur_list = new_list
                break
            else:
                # Keep filtering
                cur_list = new_list
        return cur_list


ANY_SCRIPT = object()
'''Sentinel for preference matching on any script.'''


class FontFinderException(Exception):
    '''Base Exception class for this library.'''
    pass


class UnsupportedPlatformException(FontFinderException):
    '''Raised when an operation is not supported on the current operating system.'''
    pass




'''
Hooks for pyinstaller
'''

import os


def get_hook_dirs():
    '''Returns string of parent directory of this file, which is where hook modules are located.'''
    return [os.path.dirname(__file__)]


# Code in this module derived from FindSystemFontFilename project, under the MIT Licence:
# https://github.com/moi15moi/FindSystemFontsFilename/
#
# See ACKNOWLEDGEMENTS file.
#
import platform
if platform.system() == "Windows":
    import comtypes
    from comtypes import COMError, GUID, HRESULT, IUnknown, STDMETHOD, WINFUNCTYPE
    from sys import getwindowsversion
    import winreg
    from ctypes import wintypes

    import ctypes
    from ctypes import byref, POINTER
    import os
    from pathlib import Path
    import shutil

    from fontfinder import FontFinder, FontFinderException
    import fontfinder._platforms


    USER_FONT_DIR = Path("~\\AppData\\Local\\Microsoft\\Windows\\Fonts").expanduser()
    USER_FONT_REG_PATH = "Software\\Microsoft\\Windows NT\\CurrentVersion\\Fonts"


    class WindowsPlatform(fontfinder._platforms.FontPlatform):
        def all_installed_families(self):
            dw = DirectWriteLibrary()
            dw_factory = POINTER(IDWriteFactory)()
            dw.DWriteCreateFactory(dw.DWRITE_FACTORY_TYPE_ISOLATED, IDWriteFactory._iid_, byref(dw_factory))

            fonts = POINTER(IDWriteFontCollection)()
            dw_factory.GetSystemFontCollection(byref(fonts), False)

            # Use a dict as an ordered set
            family_names = {}
            for i in range(fonts.GetFontFamilyCount()):
                font_family = POINTER(IDWriteFontFamily)()
                fonts.GetFontFamily(i, byref(font_family))

                family_name_strings = POINTER(IDWriteLocalizedStrings)()
                font_family.GetFamilyNames(byref(family_name_strings))

                name_index = 0
                name_len = ctypes.c_uint32()
                family_name_strings.GetStringLength(name_index, byref(name_len))

                name_buffer = ctypes.create_unicode_buffer(name_len.value+1)
                family_name_strings.GetString(name_index, name_buffer, name_len.value+1)
                family_names[name_buffer.value] = 1
            return list(family_names.keys())

        def install_fonts(self, font_infos):
            user32 = User32Library()
            reg_font_key = winreg.CreateKeyEx(winreg.HKEY_CURRENT_USER, USER_FONT_REG_PATH)
            for font_info in font_infos:
                dest_path = USER_FONT_DIR / font_info.filename
                font_family_subfamily = f"{font_info.family_name} {font_info.subfamily_name}".strip()
                if font_info.downloaded_path is not None and font_info.downloaded_path != Path():
                    shutil.copy2(font_info.downloaded_path, USER_FONT_DIR)
                else:
                    raise FontFinderException("Can't install font without a path to the downloaded font file")
                winreg.SetValueEx(reg_font_key, font_family_subfamily, 0, winreg.REG_SZ, str(dest_path))
            reg_font_key.Close()
            user32.SendNotifyMessageW(user32.HWND_BROADCAST, user32.WM_FONTCHANGE, 0, 0)

        def uninstall_fonts(self, font_infos):
            user32 = User32Library()
            reg_font_key = winreg.CreateKeyEx(winreg.HKEY_CURRENT_USER, USER_FONT_REG_PATH)
            for font_info in font_infos:
                if font_info.family_name is None or font_info.family_name == "":
                    raise FontFinderException("Can't uninstall font without a family name")
                winreg.DeleteValue(reg_font_key, font_info.fullname)
                if font_info.filename is None:
                    raise FontFinderException("Can't uninstall font without a filename")
                os.remove(USER_FONT_DIR / font_info.filename)
            reg_font_key.Close()
            user32.SendNotifyMessageW(user32.HWND_BROADCAST, user32.WM_FONTCHANGE, 0, 0)

        def known_platform_fonts(self) -> list[fontfinder.FontInfo]:
            # Use Segoe UI Emoji for emoji.
            font_infos = []
            font_infos.append(fontfinder.FontInfo(main_script="Common", script_variant="Emoji",
                            family_name="Segoe UI Emoji",
                            subfamily_name="Regular", postscript_name="SegoeUIEmoji",
                            form=fontfinder.FontForm.UNSET, width=fontfinder.FontWidth.NORMAL,
                            weight=fontfinder.FontWeight.REGULAR, style=fontfinder.FontStyle.UPRIGHT,
                            format=fontfinder.FontFormat.TTF, build=fontfinder.FontBuild.UNSET,
                            url=""))
            return font_infos
        
        def set_platform_prefs(self, font_finder: FontFinder) -> None:
            # Use Segoe UI Emoji for emoji.
            font_finder.family_prefs[("Common", "Emoji")] = [fontfinder.filters.attr_in("family_name",
                                                                                        ["Segoe UI Emoji"])]


    class IDWriteLocalizedStrings(IUnknown):
        _iid_ = GUID("{08256209-099a-4b34-b86d-c22b110e7771}")
        _methods_ = [
            STDMETHOD(ctypes.c_uint32, "GetCount"),
            STDMETHOD(HRESULT, "FindLocaleName"),
            STDMETHOD(HRESULT, "GetLocaleNameLength"),
            STDMETHOD(HRESULT, "GetLocaleName"),
            STDMETHOD(HRESULT, "GetStringLength", [ctypes.c_uint32, POINTER(ctypes.c_uint32)]),
            STDMETHOD(HRESULT, "GetString", [ctypes.c_uint32, POINTER(wintypes.WCHAR), ctypes.c_uint32])
        ]


    class IDWriteFontList(IUnknown):
        # https://learn.microsoft.com/en-us/windows/win32/api/dwrite/nn-dwrite-idwritefontlist
        _iid_ = GUID("{1a0d8438-1d97-4ec1-aef9-a2fb86ed6acb}")
        _methods_ = [
            STDMETHOD(None, "GetFontCollection"),       # Not implemented here
            STDMETHOD(wintypes.UINT, "GetFontCount"),
            STDMETHOD(HRESULT, "GetFont"),              # Not implemented here
        ]


    class IDWriteFontFamily(IDWriteFontList):
        # https://learn.microsoft.com/en-us/windows/win32/api/dwrite/nn-dwrite-idwritefontfamily
        _iid_ = GUID("{da20d8ef-812a-4c43-9802-62ec4abd7add}")
        _methods_ = [
            STDMETHOD(None, "GetFamilyNames", [POINTER(POINTER(IDWriteLocalizedStrings))]),
            STDMETHOD(None, "GetFirstMatchingFont"),    # Not implemented here
            STDMETHOD(None, "GetMatchingFonts"),        # Not implemented here
        ]

    class IDWriteFontCollection(IUnknown):
        # https://learn.microsoft.com/en-us/windows/win32/api/dwrite/nn-dwrite-idwritefontcollection
        _iid_ = GUID("{a84cee02-3eea-4eee-a827-87c1a02a0fcc}")
        _methods_ = [
            STDMETHOD(wintypes.UINT, "GetFontFamilyCount"),
            STDMETHOD(HRESULT, "GetFontFamily", [wintypes.UINT, POINTER(POINTER(IDWriteFontFamily))]),
            STDMETHOD(None, "FindFamilyName"),          # Not implemented here
            STDMETHOD(None, "GetFontFromFontFace"),     # Not implemented here
        ]

    class IDWriteFactory(IUnknown):
        # https://learn.microsoft.com/en-us/windows/win32/api/dwrite/nn-dwrite-idwritefactory
        _iid_ = GUID("{b859ee5a-d838-4b5b-a2e8-1adc7d93db48}")
        _methods_ = [
            STDMETHOD(HRESULT, "GetSystemFontCollection", [POINTER(POINTER(IDWriteFontCollection)), wintypes.BOOLEAN]),
            STDMETHOD(None, "CreateCustomFontCollection"),      # Not implemented here
            STDMETHOD(None, "RegisterFontCollectionLoader"),    # Not implemented here
            STDMETHOD(None, "UnregisterFontCollectionLoader"),  # Not implemented here
            STDMETHOD(None, "CreateFontFileReference"),         # Not implemented here
            STDMETHOD(None, "CreateCustomFontFileReference"),   # Not implemented here
            STDMETHOD(None, "CreateFontFace"),                  # Not implemented here
            STDMETHOD(None, "CreateRenderingParams"),           # Not implemented here
            STDMETHOD(None, "CreateMonitorRenderingParams"),    # Not implemented here
            STDMETHOD(None, "CreateCustomRenderingParams"),     # Not implemented here
            STDMETHOD(None, "RegisterFontFileLoader"),          # Not implemented here
            STDMETHOD(None, "UnregisterFontFileLoader"),        # Not implemented here
            STDMETHOD(None, "CreateTextFormat"),                # Not implemented here
            STDMETHOD(None, "CreateTypography"),                # Not implemented here
            STDMETHOD(None, "GetGdiInterop"),                   # Not implemented here
            STDMETHOD(None, "CreateTextLayout"),                # Not implemented here
            STDMETHOD(None, "CreateGdiCompatibleTextLayout"),   # Not implemented here
            STDMETHOD(None, "CreateEllipsisTrimmingSign"),      # Not implemented here
            STDMETHOD(None, "CreateTextAnalyzer"),              # Not implemented here
            STDMETHOD(None, "CreateNumberSubstitution"),        # Not implemented here
            STDMETHOD(None, "CreateGlyphRunAnalysis"),          # Not implemented here
        ]


    class DirectWriteLibrary(fontfinder._platforms.CTypesLibrary):
        def __init__(self):
            super().__init__(ctypes.windll, "dwrite")

            # I couldn't make this work using the ctypes function protype technique, so we just use the 
            # standard foreign function attribute technique.
            self.DWriteCreateFactory = self.lib.DWriteCreateFactory
            self.DWriteCreateFactory.restype = HRESULT
            self.DWriteCreateFactory.argtypes = [wintypes.UINT, GUID, POINTER(POINTER(IUnknown))]

            self.DWRITE_FACTORY_TYPE_SHARED = 0
            self.DWRITE_FACTORY_TYPE_ISOLATED = 1


    class User32Library(fontfinder._platforms.CTypesLibrary):
        def __init__(self):
            super().__init__(ctypes.windll, "User32")

            self.SendNotifyMessageW = self.w_prototype(
                wintypes.BOOL, "SendNotifyMessageW", (self.IN, wintypes.HWND, "hWnd"),
                                                     (self.IN, wintypes.UINT, "Msg"),
                                                     (self.IN, wintypes.WPARAM, "wParam"),
                                                     (self.IN, wintypes.LPARAM, "lParam")
            )

            self.HWND_BROADCAST = wintypes.HWND(0xffff)
            self.WM_FONTCHANGE = wintypes.UINT(0x001D)



'''Dataclass objects and enumerations used by `fontfinder`.
'''
from collections import Counter
import csv
import dataclasses
from dataclasses import dataclass, field
import functools
from enum import Enum, Flag, auto
from pathlib import Path, PurePosixPath
import re
from urllib.parse import urlparse


@dataclass
class TextInfo:
    '''Stories Unicode script information about a string of text, suitable for selecting the best font to display
    the text.
    '''
    main_script: str = ""
    '''Name of the most frequently used Unicode script in a piece of text. This is the long Unicode script value
    (known as a property value alias), rather than the short 4-character script code.'''

    script_variant: str = ""
    '''A secondary string used when the value of `main_script` is insufficient for choosing an appropriate font.
    This is not a Unicode property, but a scheme only used by `fontfinder`. See `FontFinder.analyse()` for examples.'''

    emoji_count: int = 0
    '''Count of characters who have either the Emoji Presentation property or the Extended_Pictographic property set
    (independent of script).'''

    script_count: Counter = field(default_factory=Counter)
    '''A [collections.Counter](https://docs.python.org/3/library/collections.html#collections.Counter) of the count
    of each Unicode script in the text. The keys are the string names of each script that appears in the text.'''


@dataclasses.dataclass(order=True)
class FontInfo:
    '''Stores fonta metadata about an individual font file.'''

    main_script: str
    '''Primary Unicode script covered by the font. The same format as `TextInfo.main_script`.'''

    script_variant: str
    '''Script variant covered by the font. The same format as `TextInfo.script_variant`.'''

    family_name: str
    '''Font family name.'''

    subfamily_name: str
    '''Font subfamily name (sometimes also referred to as the style name).'''

    postscript_name: str
    '''Font PostScript name.'''

    form: 'FontForm'
    '''Enum of the font form.'''

    width: 'FontWidth'
    '''Enum of the font width.'''

    weight: 'FontWeight'
    '''Enum of the font weight.'''

    style: 'FontStyle'
    '''Enum of the font style.'''

    format: 'FontFormat'
    '''Enum of the font file format.'''

    build: 'FontBuild'
    '''Enum of the font build (mainly for Google Noto fonts).'''

    tags: 'FontTag'
    '''Enum of extra tags describing the font file.'''

    url: str = ""
    '''URL download source for the font. Empty string if download is not available.'''

    downloaded_path: Path = Path()
    '''Path to the downloaded font on the local filesystem for installation. Empty if not set.'''

    def __init__(self, main_script: str = None, script_variant: str = None, family_name: str = None,
                 subfamily_name: str = None, postscript_name: str = None, form: 'FontForm' = None,
                 width: 'FontWidth' = None, weight: 'FontWeight' = None, style: 'FontStyle' = None,
                 format: 'FontFormat' = None, build: 'FontBuild' = None, tags: 'FontTag' = None, url: str  = None,
                 downloaded_path: Path = None 
                ):

        self.main_script = "" if main_script is None else main_script
        self.script_variant = "" if script_variant is None else script_variant
        self.family_name = "" if family_name is None else family_name
        self.subfamily_name = "" if subfamily_name is None else subfamily_name
        self.postscript_name = "" if postscript_name is None else postscript_name
        self.form = FontForm.UNSET if form is None else form
        self.width = FontWidth.NORMAL if width is None else width
        self.weight = FontWeight.REGULAR if weight is None else weight
        self.style = FontStyle.UPRIGHT if style is None else style
        self.format = FontFormat.UNSET if format is None else format
        self.build = FontBuild.UNSET if build is None else build
        self.tags = FontTag(0) if tags is None else tags
        self.url = "" if url is None else url
        self.downloaded_path = Path() if downloaded_path is None else downloaded_path

    def init_from_noto_url(self, url):
        '''Uses the url of a Google Noto font to set as much of the font metadata as possible.'''
        url_path = urlparse(url).path
        self.form = FontForm.from_str(url_path)
        self.width = FontWidth.from_str(url_path)
        self.weight = FontWeight.from_str(url_path)
        self.style = FontStyle.from_str(url_path)
        self.format = FontFormat.from_str(url_path)
        self.build = FontBuild.from_str(url_path)
        
        stem = PurePosixPath(url_path).stem
        self.postscript_name = re.sub(r"\[.*\]", "", stem)
        # Previously: self.subfamily_name = self.postscript_name.split('-')[-1]
        self.subfamily_name = " ".join([self.width.text, self.weight.text, self.style.text]).strip()

        if self.width is FontWidth.VARIABLE or self.weight is FontWeight.VARIABLE:
            self.subfamily_name = FontWeight.REGULAR.text
            self.postscript_name += "-" + FontWeight.REGULAR.text
        
        if "/slim".casefold() in url_path.casefold():
            self.tags |= FontTag.SLIM
        if "Mono".casefold() in self.postscript_name.casefold():
            self.tags |= FontTag.MONO
        if "UI" in self.postscript_name:
            self.tags |= FontTag.UI
        if "Display".casefold() in self.postscript_name.casefold():
            self.tags |= FontTag.DISPLAY
        if "Looped".casefold() in self.family_name.casefold():
            self.tags |= FontTag.LOOPED
        if self.postscript_name.startswith("NotoSansNotoSansTifinagh"):
            pass
        match = re.match(r"NotoSansTifinagh(?P<variant>.*?)-", self.postscript_name)
        if match is not None:
            self.script_variant = match['variant']
        self.url = url

    @property
    def filename(self):
        '''Returns just the filename component of this font file, using the path if it has one, or otherwise its
        URL.'''
        if self.downloaded_path is not None and self.downloaded_path != Path():
            filename = self.downloaded_path.name
        elif self.url is not None and self.url != "":
            filename = PurePosixPath(urlparse(self.url).path).name
        else:
            filename = None
        return filename 

    @property
    def fullname(self):
        '''Returns a string of the family and subfamily named combined (separated by a space).'''
        return f"{self.family_name} {self.subfamily_name}".strip()

    def copy(self):
        '''Returns a copy of this `FontInfo`.'''
        return FontInfo(**dataclasses.asdict(self))

    def str_dict(self):
        '''Returns a dictionary of strings representing this object.'''
        str_dict = dataclasses.asdict(self)
        # Convert Enum-type fields to their string names without the Enum type-name
        for field_name, field_value in str_dict.items():
            if isinstance(field_value, Flag):
                name_list = []
                for member in type(field_value):
                    if member in field_value:
                        name_list.append(member.name)
                str_dict[field_name] = "|".join(name_list)
            elif isinstance(field_value, Enum):
                str_dict[field_name] = str(field_value.name)
            elif isinstance(field_value, Path):
                if field_value == Path():
                    str_dict[field_name] = ""
                else:
                    str_dict[field_name] = str(field_value)
        return str_dict
    
    @classmethod
    def from_str_dict(cls, str_dict):
        '''Takes a dictionary of strings and returns a new FontInfo object.'''
        # Convert the string names of Enum-type fields to Enum instances
        # Use an empty (but initialized) instance to confirm actual field types. (This is because we've used forward
        # references for their type declaration, which results in their dataclasses.field type showing
        # as str, not Enum)
        blank_obj_dict = dataclasses.asdict(cls())
        for field_name, field_value in blank_obj_dict.items():
            if isinstance(field_value, Flag):
                new_value = type(field_value)(0)
                for member_name in str_dict[field_name].split("|"):
                    if len(member_name) > 0:
                        new_value |= type(field_value)[member_name]
                str_dict[field_name] = new_value
            elif isinstance(field_value, Enum):
                # Indexing an Enum with the string member name returns the member
                str_dict[field_name] = type(field_value)[str_dict[field_name]]
            elif isinstance(field_value, bool):
                str_dict[field_name] = (str_dict[field_name].upper() == "TRUE")
            elif isinstance(field_value, Path):
                print(str_dict[field_name])
                if str_dict[field_name] == "":
                    str_dict[field_name] = Path()
                else:
                    str_dict[field_name] = Path(str_dict[field_name])
        return cls(**str_dict)


def write_font_infos_to_csv(font_infos, csv_path):
    '''Write a list of `FontInfo` objects to a CSV file with the given `csv_path`.'''
    with open(csv_path, "w", encoding="utf-8") as file:
        field_names = [field.name for field in dataclasses.fields(FontInfo)]
        writer = csv.DictWriter(file, field_names)
        writer.writeheader()
        for font in font_infos:
            writer.writerow(font.str_dict())

def read_font_infos_from_csv(csv_path):
    '''Read a CSV file at `csv_path` previously created by `write_font_infos_to_csv()` and use it to create and
    return a list of `FontInfo` objects.'''
    font_infos = []
    with open(csv_path, "r", encoding="utf-8") as file:
        reader = csv.DictReader(file)
        for row in reader:
            font_infos.append(FontInfo.from_str_dict(row))
    return font_infos


@functools.total_ordering
class FontForm(Enum):
    '''Enum of font forms.'''
    UNSET       = auto()
    SERIF       = auto()
    SANS_SERIF  = auto()
    # Arabic forms
    NASKH       = auto()
    # Perso-Arabic forms
    NASTALIQ    = auto()
    # Hebrew forms
    RASHI       = auto()

    @property
    def text(self):
        return font_form_str_data[self][0]

    @classmethod
    def from_str(cls, string: str):
        result = FontForm.UNSET
        for font_form, data in font_form_str_data.items():
            if data[1].search(string):
                result = font_form
        return result
    
    def __lt__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.value < other.value


font_form_str_data = {
    FontForm.SERIF:         ("Serif",       re.compile(r"Serif",    re.IGNORECASE)),
    FontForm.SANS_SERIF:    ("Sans",        re.compile(r"Sans",     re.IGNORECASE)),
    FontForm.NASKH:         ("Naskh",       re.compile(r"Naskh",    re.IGNORECASE)),
    FontForm.NASTALIQ:      ("Nastaliq",    re.compile(r"Nastaliq", re.IGNORECASE)),
    FontForm.RASHI:         ("Rashi",       re.compile(r"Rashi",    re.IGNORECASE)),
}
'''Data for string conversion to and from `FontForm`.'''


@functools.total_ordering
class FontWidth(Enum):
    '''Enum of font widths.'''
    NORMAL      = auto()
    VARIABLE    = auto()
    EXTRA_COND  = auto()
    CONDENSED   = auto()
    SEMI_COND   = auto()

    @property
    def text(self):
        if self is FontWidth.NORMAL:
            result = ""
        else:
            result = font_width_str_data[self][0]
        return result

    @classmethod
    def from_str(cls, string: str):
        result = FontWidth.NORMAL
        for font_width, data in font_width_str_data.items():
            if data[1].search(string):
                result = font_width
        return result

    def __lt__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.value < other.value


font_width_str_data = {
    FontWidth.VARIABLE:     ("wdth",           re.compile(r"wdth",             re.IGNORECASE)),
    FontWidth.CONDENSED:    ("Condensed",      re.compile(r"Condensed",        re.IGNORECASE)),
    FontWidth.EXTRA_COND:   ("ExtraCondensed", re.compile(r"Extra.?Condensed", re.IGNORECASE)),
    FontWidth.SEMI_COND:    ("SemiCondensed",  re.compile(r"Semi.?Condensed",  re.IGNORECASE)),
}
'''Data for string conversion to and from `FontWidth`.'''


@functools.total_ordering
class FontWeight(Enum):
    '''Enum of font weights.'''
    REGULAR     = auto()
    VARIABLE    = auto()
    DEMI_LIGHT  = auto()
    EXTRA_LIGHT = auto()
    LIGHT       = auto()
    THIN        = auto()
    MEDIUM      = auto()
    SEMI_BOLD   = auto()
    BOLD        = auto()
    EXTRA_BOLD  = auto()
    BLACK       = auto()

    @property
    def text(self):
        return font_weight_str_data[self][0]

    @classmethod
    def from_str(cls, string: str):
        result = FontWeight.REGULAR
        for font_width, data in font_weight_str_data.items():
            if data[1].search(string):
                result = font_width
        return result

    def __lt__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.value < other.value


font_weight_str_data = {
    FontWeight.VARIABLE:        ("wght",       re.compile(r"wght",             re.IGNORECASE)),
    FontWeight.REGULAR:         ("Regular",    re.compile(r"Regular",          re.IGNORECASE)),
    FontWeight.LIGHT:           ("Light",      re.compile(r"Light",            re.IGNORECASE)),
    FontWeight.DEMI_LIGHT:      ("DemiLight",  re.compile(r"Demi.?Light",      re.IGNORECASE)),
    FontWeight.EXTRA_LIGHT:     ("ExtraLight", re.compile(r"Extra.?Light",     re.IGNORECASE)),
    FontWeight.THIN:            ("Thin",       re.compile(r"Thin",             re.IGNORECASE)),
    FontWeight.MEDIUM:          ("Medium",     re.compile(r"Medium",           re.IGNORECASE)),
    FontWeight.BOLD:            ("Bold",       re.compile(r"Bold",             re.IGNORECASE)),
    FontWeight.SEMI_BOLD:       ("SemiBold",   re.compile(r"Semi.?Bold",       re.IGNORECASE)),
    FontWeight.EXTRA_BOLD:      ("ExtraBold",  re.compile(r"Extra.?Bold",      re.IGNORECASE)),
    FontWeight.BLACK:           ("Black",      re.compile(r"Black",            re.IGNORECASE)),
}
'''Data for string conversion to and from `FontWeight`.'''


@functools.total_ordering
class FontStyle(Enum):
    '''Enum of font styles.'''
    UPRIGHT     = auto()
    ITALIC      = auto()
 
    @property
    def text(self):
        if self is FontStyle.UPRIGHT:
            result = ""
        else:
            result = font_style_str_data[self][0]
        return result

    @classmethod
    def from_str(cls, string: str):
        result = FontStyle.UPRIGHT
        for font_style, data in font_style_str_data.items():
            if data[1].search(string):
                result = font_style
        return result

    def __lt__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.value < other.value


font_style_str_data = {
    FontStyle.ITALIC:          ("Italic", re.compile(r"Italic",            re.IGNORECASE)),
}
'''Data for string conversion to and from `FontStyle`.'''


@functools.total_ordering
class FontFormat(Enum):
    '''Enum of font file formats.'''
    UNSET       = ""
    OTF         = "OTF"
    OTC         = "OTC"
    TTF         = "TTF"

    @property
    def text(self):
        return font_format_str_data[self][0]

    @classmethod
    def from_str(cls, string: str):
        result = FontFormat.UNSET
        for font_format, data in font_format_str_data.items():
            if data[1].search(string):
                result = font_format
        return result

    def __lt__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.value < other.value


font_format_str_data = {
    FontFormat.OTF:             ("OTF", re.compile(r"\.OTF",            re.IGNORECASE)),
    FontFormat.OTC:             ("OTC", re.compile(r"\.OTC",            re.IGNORECASE)),
    FontFormat.TTF:             ("TTF", re.compile(r"\.TTF",            re.IGNORECASE)),
}
'''Data for string conversion to and from `FontFormat`.'''


@functools.total_ordering
class FontBuild(Enum):
    '''Enum of font builds (mainly for Google Noto fonts).'''
    UNSET       = auto()
    UNHINTED    = auto()
    HINTED      = auto()
    FULL        = auto()

    @property
    def text(self):
        return font_build_str_data[self][0]

    @classmethod
    def from_str(cls, string: str):
        result = FontBuild.UNSET
        for font_build, data in font_build_str_data.items():
            if data[1].search(string):
                result = font_build
        return result

    def __lt__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.value < other.value


font_build_str_data = {
    FontBuild.HINTED:      ("Hinted",   re.compile(r"Hinted",    re.IGNORECASE)),
    FontBuild.UNHINTED:    ("Unhinted", re.compile(r"Unhinted",  re.IGNORECASE)),
    FontBuild.FULL:        ("Full",     re.compile(r"Full",      re.IGNORECASE))
}
'''Data for string conversion to and from `FontBuild`.'''


@functools.total_ordering
class FontTag(Flag):
    '''Enum of extra tags describing the font file.'''
    MONO        = auto()
    '''A monospaced font.'''
    UI          = auto()
    '''A font aimed at UI display.'''
    DISPLAY     = auto()
    '''A display font.'''
    SLIM        = auto()
    '''A Google Noto slim-build variable font.'''
    LOOPED      = auto()
    '''A looped variant (e.g. for Thai fonts)'''

    def __lt__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.value < other.value



import copy
import datetime
import json
from pathlib import Path
import requests
import shutil

import fontfinder
from fontfinder.model import FontInfo, FontForm, FontWidth, FontWeight, FontStyle, FontFormat, FontBuild 


NOTO_MAIN_JSON_URL = "https://notofonts.github.io/noto.json"
'''URL of main Noto JSON font data.'''

NOTO_MAIN_BASE_URL = "https://cdn.jsdelivr.net/gh/notofonts/notofonts.github.io/"
'''Base URL of main Noto font download location.'''

NOTO_CJK_BASE_URL = "https://github.com/notofonts/noto-cjk/raw/main/"
'''Base URL of CJK Noto font download location.'''


_NOTO_MAIN_JSON_REF_PATH = Path(fontfinder._REF_DATA_DIR_PATH, "noto.json").resolve()
'''Path of reference copy of noto.json distributed with this package.'''

_NOTO_MAIN_JSON_USER_PATH = Path(fontfinder._USER_DATA_DIR_PATH, "cache", "noto.json").resolve()
'''Path of updated, cached copy of noto.json.'''

_NOTO_MAIN_JSON_MAX_AGE = datetime.timedelta(days=1)
'''Max age of cached copy of noto.json, after which an updated copy will be downloaded.'''


def get_noto_fonts(filter_func = None):
    '''Return a list of FontInfo records for the Google Noto fonts.'''
    font_infos = []
    font_infos.extend(_get_noto_main_fonts(filter_func))
    font_infos.extend(_get_noto_cjk_fonts(filter_func))
    font_infos.extend(_get_noto_emoji_fonts(filter_func))
    font_infos.sort()
    return font_infos

def _get_noto_main_data():
    '''Return main Noto JSON data as a Python object, handling cache as necessary.'''
    if not _NOTO_MAIN_JSON_USER_PATH.exists():
        # Copy noto.json distributed with this package
        _NOTO_MAIN_JSON_USER_PATH.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy2(_NOTO_MAIN_JSON_REF_PATH, _NOTO_MAIN_JSON_USER_PATH)

    last_mod_time = datetime.datetime.fromtimestamp(_NOTO_MAIN_JSON_USER_PATH.stat().st_mtime)
    if (datetime.datetime.now() - last_mod_time) >= _NOTO_MAIN_JSON_MAX_AGE:
        # Update cached noto.json
        noto_json_text = requests.get(NOTO_MAIN_JSON_URL)
        with open(_NOTO_MAIN_JSON_USER_PATH, "w", encoding="utf-8") as file:
            file.write(noto_json_text.text)
    
    # Read cached noto.json
    with open(_NOTO_MAIN_JSON_USER_PATH, "r", encoding="utf-8") as file:
        noto_data = json.load(file)
    return noto_data
            
def _get_noto_main_fonts(filter_func = None):
    '''Return a list of FontInfo records for the main (non-CJK) Google Noto fonts.'''
    font_infos = []
    noto_data = _get_noto_main_data()

    # Note that the script keys in the Noto JSON data are *mostly* Unicode script names (once they are
    # changed to Titlecase and have hyphens replaced with underscores). But some of them are "pseudo-script-names"
    # for characters that formally belong under other Unicode script names. We adjust for all this below.
    for raw_script_key, script_data in noto_data.items():
        if raw_script_key == "latin-greek-cyrillic":
            # The Noto data treats these 3 scripts as one, but we duplicate the font info for all 3 indiv scripts.
            script_key_set = ['Latin', 'Greek', 'Cyrillic']
        elif raw_script_key == "meroitic":
            # The Noto data treats Meroitic as a single script, but in the Unicode data it's two separate scripts.
            script_key_set = ['Meroitic_Cursive', 'Meroitic_Hieroglyphs']
        else:
            script_key_set = [raw_script_key]

        for script_key in script_key_set:
            script_variant = ""
            if script_key == "sign-writing":
                main_script = "SignWriting" # No hyphen or underscore in the Unicode script name
            elif script_key == "nastaliq":
                # This is a Noto "pseudo-script-name" that actually is a Nastaliq Urdu variant of Arabic script.
                main_script = "Arabic"
                script_variant = "Urdu"
            elif script_key == "math" or script_key == "music" or script_key == "symbols" or \
                 script_key == "mayan-numerals" or script_key == "indic-siyaq-numbers" or \
                 script_key == "ottoman-siyaq-numbers" or script_key == "znamenny":
                # These are all "pseudo-script-names" for characters who actual script is mostly "Common".
                main_script = "Common"
                script_variant = script_key.replace('-', '_').title()
            elif script_key == "test":
                # For completeness, we include the Noto Sans Test and Noto Serif Test fonts, but without including
                # a Unicode script.
                main_script = ""
                script_variant = "Test"
            else:
                # Make the Noto script key formatting match the Unicode script name formatting.
                main_script = script_key.replace('-', '_').title()

            for family_name, family_data in script_data['families'].items():
                if family_name == "Noto Sans Symbols2":
                    family_name = "Noto Sans Symbols 2" # Fix spacing in Noto family name
                if "Syriac Western" in family_name:
                    script_variant = "Western"
                elif "Syriac Eastern" in family_name:
                    script_variant = "Eastern"
                
                form = FontForm.from_str(family_name)

                # Some font families should be added under other scripts as well. We add them here,
                # then re-iterate over the expanded script and variant names. (Most of the time this is a single
                # iteration that changes nothing.)
                expanded_scripts = [(main_script, script_variant)]
                if family_name == "Noto Sans Symbols 2":
                    # Only the Noto Sans Symbols 2 family handles the Unicode Braille script.
                    expanded_scripts.append(("Braille", ""))

                for main_script, script_variant in expanded_scripts:
                    for build, relative_url_list in family_data['files'].items():
                        build = FontBuild.from_str(build)
                        for relative_url in relative_url_list:
                            url = NOTO_MAIN_BASE_URL + relative_url
                            font_info = FontInfo(main_script=main_script, script_variant=script_variant,
                                                 family_name=family_name, url=url)
                            font_info.init_from_noto_url(url)
                            # Form and build have already been set from the URL, but we can ensure the values are
                            # correct from the other JSON data.
                            font_info.form = form
                            font_info.build = build
                            if filter_func is None or filter_func(font_info):
                                font_infos.append(font_info)
    return font_infos


_CJK_WEIGHTS =   [
    ("Black",       FontWeight.BLACK),
    ("Bold",        FontWeight.BOLD),
    ("DemiLight",   FontWeight.DEMI_LIGHT),
    ("Light",       FontWeight.LIGHT),
    ("Medium",      FontWeight.MEDIUM),
    ("Regular",     FontWeight.REGULAR),
    ("Thin",        FontWeight.THIN),                
]
'''Available weights of Noto CJK fonts.'''

# Keys for _CJK_DATA
_CJK_SCRIPT_INFO_KEY = "script_info"
_CJK_URL_COMPONENT_KEY = "url_component"
_CJK_CODE_KEY = "cjk_code"

_CJK_DATA = {
    "chinese-simplified": {
            _CJK_SCRIPT_INFO_KEY:   [("Han", "zh-Hans")],
            _CJK_URL_COMPONENT_KEY: "SimplifiedChinese/",
            _CJK_CODE_KEY:          "SC"
        },
    "chinese-traditional": {
            _CJK_SCRIPT_INFO_KEY:   [("Han", "zh-Hant"), ("Bopomofo", "")],
            _CJK_URL_COMPONENT_KEY: "TraditionalChinese/",
            _CJK_CODE_KEY:          "TC",
        },
    "chinese-hongkong": {
            _CJK_SCRIPT_INFO_KEY:   [("Han", "zh-Hant-HK")],
            _CJK_URL_COMPONENT_KEY: "TraditionalChineseHK/",
            _CJK_CODE_KEY:          "HK",
        },
    "japanese": {
            _CJK_SCRIPT_INFO_KEY:   [("Katakana", ""), ("Hiragana", ""), ("Katakana_Or_Hiragana", ""), ("Han", "ja")],
            _CJK_URL_COMPONENT_KEY: "Japanese/",
            _CJK_CODE_KEY:          "JP",
        },
    "korean": {
            _CJK_SCRIPT_INFO_KEY:   [("Hangul", ""), ("Han", "ko")],
            _CJK_URL_COMPONENT_KEY: "Korean/",
            _CJK_CODE_KEY:          "KR",
        },
}
'''Data needed to create FontInfo records for Noto CJK fonts.'''


def _get_noto_cjk_fonts(filter_func = None):
    '''Return a list of FontInfo records for the CJK Google Noto fonts.'''
    font_infos = []
    for lang, lang_data in _CJK_DATA.items():
        cjk_code = lang_data[_CJK_CODE_KEY]
        url_component = lang_data[_CJK_URL_COMPONENT_KEY]
        script_infos = lang_data[_CJK_SCRIPT_INFO_KEY]
        for form in [FontForm.SANS_SERIF, FontForm.SERIF]:
            form_name = "Sans" if form is FontForm.SANS_SERIF else "Serif"
            lang_fonts = []
            for weight_name, weight in _CJK_WEIGHTS:
                # We're using the language-specific OTF versions of the Noto CJK fonts.
                family_name = f"Noto {form_name} CJK {cjk_code.upper()}"
                postscript_name = f"Noto{form_name}CJK{cjk_code.lower()}-{weight_name}"
                url = f"{NOTO_CJK_BASE_URL}{form_name}/OTF/{url_component}{postscript_name}.otf"
                lang_fonts.append(
                    FontInfo(main_script="", script_variant="", family_name=family_name, subfamily_name=weight_name,
                            postscript_name=postscript_name, form=form, width=FontWidth.NORMAL,
                            weight=weight, style=FontStyle.UPRIGHT, format=FontFormat.OTF,
                            build=FontBuild.FULL, url=url
                            )                
                )
            
            for main_script, script_variant in script_infos:
                for font_info in lang_fonts:
                    font_info = copy.copy(font_info)
                    font_info.main_script = main_script
                    font_info.script_variant = script_variant
                    if filter_func is None or filter_func(font_info):
                        font_infos.append(font_info)
    return font_infos

def _get_noto_emoji_fonts(filter_func = None):
    '''Return a list of FontInfo records for Google Noto emoji fonts. Currently just Noto Color Emoji.'''
    font_infos = []
    font_info = FontInfo(main_script="Common", script_variant="Emoji", family_name="Noto Color Emoji",
                        subfamily_name="Regular", postscript_name="NotoColorEmoji",
                        form=FontForm.UNSET, width=FontWidth.NORMAL, weight=FontWeight.REGULAR,
                        style=FontStyle.UPRIGHT, format=FontFormat.TTF, build=FontBuild.UNSET,
                        # Note: This url may not be the fully-built version of the font:
                        url="https://github.com/googlefonts/noto-emoji/raw/main/fonts/NotoColorEmoji.ttf")
    if filter_func is None or filter_func(font_info):
        font_infos.append(font_info)
    return font_infos
    



from collections import Counter
import contextlib
import filecmp
import platform
from pprint import pprint
import tempfile
import time

import unicodedataplus as udp
import pytest

from fontfinder import *


FONT_INSTALL_SLEEP_STEP = 1
MAX_FONT_INSTALL_RETRIES = 20


class TestMode(Enum):
    TEST    = auto()    # Run the test as normal
    CREATE  = auto()    # Create the expected output, rather than testing for it
    OBSERVE = auto()    # Observe output: Use known instead of temp directories, don't unzip word docs, don't test


class FontFinderWithTestFonts(FontFinder):
    def get_test_font_infos(self):
        font_infos = []
        font_infos.append(FontInfo(main_script="Latin", script_variant="Fontfinder", family_name="Fontfinder",
                          subfamily_name="Regular", postscript_name="Fontfinder-Regular", form=FontForm.SANS_SERIF,
                          width=FontWidth.NORMAL, weight=FontWeight.REGULAR, style=FontStyle.UPRIGHT,
                          format=FontFormat.OTF,
                          url="https://github.com/multiscript/fontfinder/raw/main/test/data/Fontfinder-Regular.otf",
                          downloaded_path=Path(__file__, "../data/Fontfinder-Regular.otf").resolve())
                         )
        font_infos.append(FontInfo(main_script="Latin", script_variant="Fontfinder", family_name="Fontfinder",
                          subfamily_name="Bold", postscript_name="Fontfinder-Bold", form=FontForm.SANS_SERIF,
                          width=FontWidth.NORMAL, weight=FontWeight.BOLD, style=FontStyle.UPRIGHT,
                          format=FontFormat.OTF,
                          url="https://github.com/multiscript/fontfinder/raw/main/test/data/Fontfinder-Bold.otf",
                          downloaded_path=Path(__file__, "../data/Fontfinder-Bold.otf").resolve())
                         )
        font_infos.append(FontInfo(main_script="Latin", script_variant="Fontfinder", family_name="Fontfinder",
                          subfamily_name="Italic", postscript_name="Fontfinder-Italic", form=FontForm.SANS_SERIF,
                          width=FontWidth.NORMAL, weight=FontWeight.REGULAR, style=FontStyle.ITALIC,
                          format=FontFormat.OTF,
                          url="https://github.com/multiscript/fontfinder/raw/main/test/data/Fontfinder-Italic.otf",
                          downloaded_path=Path(__file__, "../data/Fontfinder-Italic.otf").resolve())
                         )
        font_infos.append(FontInfo(main_script="Latin", script_variant="Fontfinder", family_name="Fontfinder",
                          subfamily_name="Bold Italic", postscript_name="Fontfinder-BoldItalic",
                          form=FontForm.SANS_SERIF, width=FontWidth.NORMAL, weight=FontWeight.BOLD,
                          style=FontStyle.ITALIC, format=FontFormat.OTF,
                          url="https://github.com/multiscript/fontfinder/raw/main/test/data/Fontfinder-BoldItalic.otf",
                          downloaded_path=Path(__file__, "../data/Fontfinder-BoldItalic.otf").resolve())
                         )
        return font_infos
    
    def known_fonts(self, filter_func = None):
        if self._all_known_fonts is None:
            # Add our test fonts to the cached collection
            super().known_fonts(filter_func)
            self._all_known_fonts.extend(self.get_test_font_infos())
        return super().known_fonts(filter_func)


class TestFontFinder:
    def get_test_font_info(self):
        font_info = FontInfo(main_script="Latin", family_name="Fontfinder", subfamily_name="Regular",
                             postscript_name="Fontfinder-Regular", form=FontForm.SANS_SERIF,
                             width=FontWidth.NORMAL, weight=FontWeight.REGULAR, style=FontStyle.UPRIGHT,
                             format=FontFormat.OTF,
                             url="https://github.com/multiscript/fontfinder/raw/main/test/data/Fontfinder-Regular.otf",
                             downloaded_path=Path(__file__, "../data/Fontfinder-Regular.otf").resolve()
                             )
        return font_info

    def test_attr_contains_str(self):
        filter = attr_contains_str("family_name", ["mono", "display"])
        assert filter(FontInfo(family_name="Some Font Mono"))
        assert filter(FontInfo(family_name="Some Display Font"))
        assert not filter(FontInfo(family_name="Some Font UI"))

    def test_attr_not_contains_str(self):
        filter = attr_not_contains_str("family_name", ["mono", "display"])
        assert not filter(FontInfo(family_name="Some Font Mono"))
        assert not filter(FontInfo(family_name="Some Display Font"))
        assert filter(FontInfo(family_name="Some Font UI"))

    def test_known_fonts(self):
        ff = FontFinder()
        font_infos = ff.known_fonts() # Ensure no errors in creating list
        # font_infos = [font for font in font_infos if font.family_name == "Noto Sans"]
        # print(len(fonts))
        # pprint(fonts[-10:])

    def test_known_scripts(self):
        ff = FontFinder()
        pprint(ff.known_scripts())

    def test_known_script_variants(self):
        ff = FontFinder()
        script_variants = ff.known_script_variants() # Ensure no errors
    
    def test_all_unicode_scripts(self):
        ff = FontFinder()
        pprint(ff.all_unicode_scripts())

    def test_scripts_not_known(self):
        ff = FontFinder()
        print("Unicode Scripts Not Covered:")
        scripts_not_known = ff.scripts_not_known()
        pprint(scripts_not_known)
        scripts_not_known == ['Garay', 'Gurung_Khema', 'Kirat_Rai', 'Ol_Onal', 'Sunuwar', 'Todhri', 'Tulu_Tigalari']
                             # These are new scripts in Unicode 16, not yet covered by Noto fonts.
        print("Noto Pseudo-Scripts Not in Unicode:")
        not_in_unicode = list(set(ff.known_scripts()) - set(ff.all_unicode_scripts()) - {''})
        print(not_in_unicode)
        assert len(not_in_unicode) == 0

    def test_analyse(self):
        ff = FontFinder()
        for sample_text in sample_texts:
            text_info = ff.analyse(sample_text['text'])
            assert sample_text['main_script'] == text_info.main_script
            assert sample_text['script_variant'] == text_info.script_variant

    def test_empty_text(self):
        ff = FontFinder()
        text_info = ff.analyse('')
        assert text_info.main_script == ''
        assert text_info.script_variant == ''
        assert text_info.emoji_count == 0
        assert len(text_info.script_count) == 0

        assert ff.find_families(text_info) == []
        assert ff.find_family(text_info) == None

        assert ff.find_families('') == []
        assert ff.find_family('') == None

    def test_find_families(self):
        ff = FontFinder()
        for sample_text in sample_texts:
            family_names = ff.find_families(sample_text['text'])
            assert sample_text['expected_family_names'] == family_names

    def test_find_family(self):
        ff = FontFinder()
        for sample_text in sample_texts:
            family_name = ff.find_family(sample_text['text'])
            assert sample_text['expected_family_name'] == family_name

    def test_find_family_fonts(self):
        ff = FontFinder()
        print("Finding family members")
        font_infos = ff.find_family_fonts("Noto Naskh Arabic")
        fullnames = {font_info.fullname for font_info in font_infos}
        assert fullnames == {'Noto Naskh Arabic Bold', 'Noto Naskh Arabic Medium',
                            'Noto Naskh Arabic Regular', 'Noto Naskh Arabic SemiBold'}
    
    def test_find_multiple_family_fonts(self):
        ff = FontFinder()
        family_1 = "Noto Naskh Arabic"
        family_2 = "Noto Sans Cherokee"

        font_infos_1 = ff.find_family_fonts(family_1)
        fullnames_1 = {font_info.fullname for font_info in font_infos_1}
        assert len(fullnames_1) > 0

        font_infos_2 = ff.find_family_fonts(family_2)
        fullnames_2 = {font_info.fullname for font_info in font_infos_2}
        assert len(fullnames_2) > 0
        assert fullnames_1 != fullnames_2

        font_infos_combo = ff.find_family_fonts([family_1, family_2])
        fullnames_combo = {font_info.fullname for font_info in font_infos_combo}
        assert fullnames_combo == fullnames_1 | fullnames_2

    def test_find_empty_family_fonts(self):
        ff = FontFinder()
        font_infos = ff.find_family_fonts([])
        assert font_infos == []

        font_infos = ff.find_family_fonts([], "Latin")
        assert font_infos == []

    def test_all_installed_families(self):
        ff = FontFinder()
        all_installed_families = ff.all_installed_families()
        pprint(all_installed_families)

    def test_known_fonts_to_csv(self, test_mode = TestMode.TEST):
        ff = FontFinder()
        font_infos = ff.known_fonts()
        filename = "known_fonts.csv"
        self._font_infos_test_to_csv(font_infos, filename, test_mode)

    def test_script_variants_to_csv(self, test_mode = TestMode.TEST):
        ff = FontFinder()
        font_infos = []
        for (main_script, script_variant) in ff.known_script_variants():
            print(f"{main_script}, {script_variant}")
            font_family = ff.find_family(TextInfo(main_script, script_variant))
            print(font_family)
            font_infos.extend(ff.find_family_fonts(font_family, main_script, script_variant))
        filename = "known_script_variants.csv"
        self._font_infos_test_to_csv(font_infos, filename, test_mode)

    def _font_infos_test_to_csv(self, font_infos, filename, test_mode):
        if test_mode is TestMode.TEST:
            context_manager = tempfile.TemporaryDirectory()
        elif test_mode is TestMode.OBSERVE:
            dir_path = Path("~/Desktop/TestFontFinder/").expanduser()
            dir_path.mkdir(parents=True, exist_ok=True)
            context_manager = contextlib.nullcontext(dir_path)
        else:
            assert False

        write_fonts = font_infos

        with context_manager as output_dir:
            csv_path = Path(output_dir, filename)
            write_font_infos_to_csv(write_fonts, csv_path)
            read_fonts = read_font_infos_from_csv(csv_path)
        
        # for item in zip(write_fonts, read_fonts):
        #     if item[0] != item[1]:
        #         print(item[0])
        #         print(item[1])
        #         assert False
        assert write_fonts == read_fonts

    def test_download_fonts(self, test_mode = TestMode.TEST):
        ff = FontFinderWithTestFonts()
        test_font_infos = ff.get_test_font_infos()
        temp_dir = None
        if test_mode is TestMode.TEST:
            temp_dir = tempfile.TemporaryDirectory()
            download_dir = temp_dir.name
        elif test_mode is TestMode.OBSERVE:
            download_dir = Path("~/Desktop/TestFontFinder/Download").expanduser()
            download_dir.mkdir(parents=True, exist_ok=True)

        ff.download_fonts(test_font_infos, download_dir)
        filenames = [font_info.filename for font_info in test_font_infos]
        cmp_result = filecmp.cmpfiles(download_dir, Path(__file__, "../data").resolve(), filenames,
                                      shallow=False)
        if temp_dir is not None:
            temp_dir.cleanup()
        assert cmp_result[1] == []
        assert cmp_result[2] == []

    def test_install_fonts(self):
        print("Uninstalling fonts")
        ff = FontFinderWithTestFonts()
        test_font_infos = ff.get_test_font_infos()
        self.uninstall_fonts_and_verify(ff, test_font_infos)

        print("Installing fonts")
        ff = FontFinderWithTestFonts()
        test_font_infos = ff.get_test_font_infos()
        self.install_fonts_and_verify(ff, test_font_infos)

        print("Uninstalling fonts")
        ff = FontFinderWithTestFonts()
        test_font_infos = ff.get_test_font_infos()
        self.uninstall_fonts_and_verify(ff, test_font_infos)

    def test_full_test(self):
        ff = FontFinderWithTestFonts()
        self.uninstall_fonts_and_verify(ff, ff.get_test_font_infos())

        text_info = TextInfo("Latin", "Fontfinder")
        font_family = ff.find_family(text_info)
        assert font_family == "Fontfinder"

        font_infos = ff.find_family_fonts(font_family)
        assert font_infos == ff.get_test_font_infos()

        font_infos = ff.find_family_fonts_to_download(font_family)
        assert font_infos == ff.get_test_font_infos()

        temp_dir = tempfile.TemporaryDirectory()
        font_infos = ff.download_fonts(font_infos, temp_dir.name)
        self.install_fonts_and_verify(ff, font_infos)
        temp_dir.cleanup()

        expect_empty = ff.find_family_fonts_to_download(font_family)
        assert len(expect_empty) == 0

        self.uninstall_fonts_and_verify(ff, font_infos)

    def test_is_rtl(self):
        ff = FontFinder()
        assert not ff.is_rtl("Latin")
        assert ff.is_rtl("Arabic")
        assert ff.is_rtl("Hebrew")

        for sample_text in sample_texts:
            text_info = ff.analyse(sample_text['text'])
            assert sample_text['is_rtl'] == ff.is_rtl(text_info)
    
    def test_is_rtl_empty_text(self):
        ff = FontFinder()
        assert ff.is_rtl('blah') == False
        assert ff.is_rtl(ff.analyse('')) == False

    def install_fonts_and_verify(self, font_finder, font_infos):
        font_finder.install_fonts(font_infos)
        for i in range(MAX_FONT_INSTALL_RETRIES):
            time.sleep(FONT_INSTALL_SLEEP_STEP)
            font_families = font_finder.all_installed_families()
            if font_infos[0].family_name in font_families:
                break
        assert font_infos[0].family_name in font_families

    def uninstall_fonts_and_verify(self, font_finder, font_infos):
        try:
            font_finder.uninstall_fonts(font_infos)
        except FileNotFoundError:
            pass
        for i in range(MAX_FONT_INSTALL_RETRIES):
            time.sleep(FONT_INSTALL_SLEEP_STEP)
            font_families = font_finder.all_installed_families()
            if font_infos[0].family_name not in font_families:
                break
        assert font_infos[0].family_name not in font_families

    @pytest.mark.skip("Investigation test to examine variants with multiple families")
    def test_multi_family_script_variants(self):
        ff = FontFinder()
        for (main_script, script_variant) in ff.known_script_variants():
            font_families = ff.find_families(TextInfo(main_script, script_variant))
            if len(font_families) > 1:
                print(f"{main_script}, {script_variant}:")
                print(font_families)
                print

    @pytest.mark.skip("Investigation test to examine script of emoji codepoints")
    def test_script_of_emoji(self):
        script_counter = Counter()
        for i in range(0x10ffff):
            char = chr(i)
            if udp.is_emoji_presentation(char) or udp.is_extended_pictographic(char):
                script_counter[udp.script(char)] += 1
        print(script_counter)

#
# These sample texts are taken from the Wikipedia article for 'Earth' in various languages.
#
sample_texts = [
{'language': 'English',
 'main_script': 'Latin',
 'script_variant': '',
 'expected_family_names': ['Noto Sans', 'Noto Sans Mono', 'Noto Serif', 'Noto Serif Display'],
 'expected_family_name': 'Noto Sans',
 'is_rtl': False,
 'text':
'''
Earth is the third planet from the Sun and the only place known in the universe where life has originated and found
habitability. While Earth may not contain the largest volumes of water in the Solar System, only Earth sustains
liquid surface water, extending over 70.8% of the planet with its ocean, making it an ocean world. The polar
regions currently retain most of all other water with large sheets of ice covering ocean and land, dwarfing Earth's
groundwater, lakes, rivers and atmospheric water. The other 29.2% of the Earth's surface is land, consisting of
continents and islands, and is widely covered by vegetation. Below the planet's surface lies the crust, consisting
of several slowly moving tectonic plates, which interact to produce mountain ranges, volcanoes, and earthquakes.
Inside the Earth's crust is a liquid outer core that generates the magnetosphere, deflecting most of the
destructive solar winds and cosmic radiation.

Earth has a dynamic atmosphere, which sustains Earth's surface conditions and protects it from most meteoroids and
UV-light at entry. It has a composition of primarily nitrogen and oxygen. Water vapor is widely present in the
atmosphere, forming clouds that cover most of the planet. The water vapor acts as a greenhouse gas and, together
with other greenhouse gases in the atmosphere, particularly carbon dioxide (CO2), creates the conditions for both
liquid surface water and water vapor to persist via the capturing of energy from the Sun's light. This process
maintains the current average surface temperature of 14.76 °C, at which water is liquid under atmospheric pressure.
Differences in the amount of captured energy between geographic regions (as with the equatorial region receiving
more sunlight than the polar regions) drive atmospheric and ocean currents, producing a global climate system with
different climate regions, and a range of weather phenomena such as precipitation, allowing components such as
nitrogen to cycle.

Earth is rounded into an ellipsoid with a circumference of about 40,000 km. It is the densest planet in the Solar
System. Of the four rocky planets, it is the largest and most massive. Earth is about eight light-minutes away from
the Sun and orbits it, taking a year (about 365.25 days) to complete one revolution. The Earth rotates around its
own axis in slightly less than a day (in about 23 hours and 56 minutes). The Earth's axis of rotation is tilted
with respect to the perpendicular to its orbital plane around the Sun, producing seasons. Earth is orbited by one
permanent natural satellite, the Moon, which orbits Earth at 384,400 km (1.28 light seconds) and is roughly a
quarter as wide as Earth. Through tidal locking, the Moon always faces the Earth with the same side, which causes
tides, stabilizes Earth's axis, and gradually slows its rotation.

Earth, like most other bodies in the Solar System, formed 4.5 billion years ago from gas in the early Solar System.
During the first billion years of Earth's history, the ocean formed and then life developed within it. Life spread
globally and has been altering Earth's atmosphere and surface, leading to the Great Oxidation Event two billion
years ago. Humans emerged 300,000 years ago, and reached a population of 8 billion on November 15, 2022. Humans
depend on Earth's biosphere and natural resources for their survival, but have increasingly impacted the planet's
environment. Humanity's current impact on Earth's climate and biosphere is unsustainable, threatening the
livelihood of humans and many other forms of life, and causing widespread extinctions.[24]
'''},

{'language': 'Chinese (Simplified)',
 'main_script': 'Han',
 'script_variant': 'zh-Hans',
 'expected_family_names': ['Noto Sans CJK SC', 'Noto Serif CJK SC'],
 'expected_family_name': 'Noto Sans CJK SC',
 'is_rtl': False,
 'text':
'''
地球是太阳系中由內及外的第三顆行星，距离太阳149 597 870.7公里/1天文單位，是宇宙中人類已知唯一存在生命的天体[3]，也
是人類居住的星球，共有80億人口[22]。其質量约为5.97×1024公斤，半径约6,371公里，平均密度5.5 g/cm3，是太阳系行星中最高
的。地球同时进行自转和公转运动，分别产生了昼夜及四季的变化更替，一太陽日自转一周，一太陽年公转一周。自转轨道面称为
赤道面，公转轨道面称为黄道面，两者之间的夹角称为黄赤交角。地球仅有一顆自然卫星，即月球。

地球表面有71%的面积被水覆盖，称为海洋或湖或河流[23][24]，其余是陆地板块組成的大洲和岛屿，表面分布河流和湖泊等水源。
南极的冰盖及北极存有冰。主體包括岩石圈、地幔、熔融态金属的外地核以及固态金属的內地核。擁有由外地核產生的地磁场
[25]。外部被氣體包圍，称为大氣層，主要成分為氮、氧、二氧化碳、氬。

地球诞生于约45.4亿年前[26][27][28][29]，42億年前開始形成海洋[30][31]，并在35亿年前的海洋中出现生命
[32][33][34][35][36]，之后逐步涉足地表和大气，并分化为好氧生物和厌氧生物。早期生命迹象产生的具體证据包括格陵兰岛西
南部变质沉积岩中拥有约37亿年的历史的生源石墨，以及澳大利亚大陆西部岩石中约41亿年前的早期生物遗骸[37][38]。此后除去
数次生物集群灭绝事件，生物种类不断增多[39]。根据科学界测定，地球曾存在过的50亿种物种中[40]，已经绝灭的占约
99%[41][42]，据统计，现今存活的物种大约有1,200至1,400万个[43][44]，其中有记录证实存活的物种120万个，而余下的86%尚未
被正式发现[45]。2016年5月，有科学家认为现今地球上大概共出现过1万亿种物种，其中人类正式发现的仅占十万分之一
[46]。2016年7月，研究团队在研究现存生物的基因后推断所有现存生物的共祖中共存在有355种基因[47][48]。地球上有约80.3亿
人口[49]，分成了约200个国家和地区，藉由外交、旅游、贸易、传媒或战争相互联系[50]。
'''},

{'Language': 'Cantonese',
 'main_script': 'Han',
 'script_variant': 'zh-Hant',
 'expected_family_names': ['Noto Sans CJK TC', 'Noto Serif CJK TC'],
 'expected_family_name': 'Noto Sans CJK TC',
 'is_rtl': False,
 'text':
'''
佢距離太陽 1.5 億公里（1個天文單位）遠，係太陽系嘅行星入面第三近太陽嘅－排正喺水星同金星之後。佢嘅質量係 5.97 ×
1024 公斤左右，半徑大約係 6371 公里，密度係每立方厘米 5.514 克。如果齋由大細嚟睇，佢喺太陽系嘅行星入面排第五，係太
陽系嘅類地行星（Terrestrial planet；指主要由石頭構成嘅行星）入面最大粒嘅；同時佢係太陽系入面密度最高嘅行星[2]。同其
他行星一樣，地球會自轉同公轉：佢大約每廿四個鐘頭會自轉一個圈（為之一日），每 365.26 日會圍住太陽公轉一個圈（為之一
個太陽年）。佢嘅自轉產生咗晝夜，而公轉就產生咗一年嘅四季。同時，地球有粒自然衛星－月球－圍住佢轉[2]。

地球嘅表面有 71% 嘅面積俾水𢫏住－呢啲面積包括咗海、河、同埋湖呀噉[2][3]，淨低嘅係由陸地板塊（Tectonic plate）組成嘅
各個大陸同埋島。兩極地區大部份地方長年俾冰覆蓋。地球嘅內部分做外地核同內地核兩層，前者由熔化咗嘅金屬組成，而後者由
固體嘅金屬組成[4]。喺地球嘅最外層，有一浸氣體包住－即係所謂嘅大氣層。大氣層嘅主要成分係氮（Nitrogen）同氧（Oxygen）
呢兩種元素。

地球係目前成個宇宙入面已知嘅行星當中唯一一個有生命存在嘅，而且佢仲係智人（Homo sapien）嘅屋企。佢喺大約 45 億 4 千
萬年前形成[2][5]，而月球就係喺之後嘅 1 千萬年開始圍住佢轉。地球形成咗 10 億年之後左右開始有生命體[6][7]，而隨住生物
嘅進化[8][9]，佢今日已經住咗各種各樣嘅生物，有各種嘅動物同植物－即係有生物多樣性（Biodiversity）[8][10]。

根據對化石同基因嘅研究，現代嘅人類喺大概 500 萬至 700 萬年之前出現，而且同現代嘅黑猩猩係近親。到咗廿一世紀初，地球
上大約有 74 億人口[11]。聯合國估話會愈嚟愈多。人類有得分做黃種人、白人、同埋黑人等多個人種，而呢柞人種入面又有啲亞
種[9]－例如係黃種人，就有得分做漢族同大和族等，而漢族仲有得細分類做包括粵人在內嘅各個民系－人類喺文化同血統上都好多
樣化。
'''},

{'language': 'Arabic',
 'main_script': 'Arabic',
 'script_variant': '',
 'expected_family_names': ['Noto Kufi Arabic', 'Noto Naskh Arabic', 'Noto Naskh Arabic UI', 'Noto Sans Arabic',
                           'Noto Sans Arabic UI'],
 'expected_family_name': 'Noto Naskh Arabic',
 'is_rtl': True,
 'text':
'''
الأَرْض (رمزها: 🜨) هي ثالث كواكب المجموعة الشمسية بعدًا عن الشمس بعد عطارد والزهرة، وتُعتبر من أكبر الكواكب
الأرضية وخامس أكبر الكواكب في النظام الشمسي، وذلك من حيث قطرها وكتلتها وكثافتها، ويُطلق على هذا الكوكب أيضًا اسم
العالم.

تعتبر الأرض مسكنًا لملايين الأنواع  من الكائنات الحية، بما فيها الإنسان؛ وهي المكان الوحيد المعروف بوجود حياة عليه
في الكون. تكونت الأرض منذ حوالي 4.54 مليار سنة، وقد ظهرت الحياة على سطحها في المليار سنة الأخيرة. ومنذ ذلك الحين
أدى الغلاف الحيوي للأرض إلى تغير الغلاف الجوي والظروف غير الحيوية الموجودة على الكوكب، مما سمح بتكاثر الكائنات التي
تعيش فقط في ظل وجود الأكسجين وتكوّن طبقة الأوزون، التي تعمل مع المجال المغناطيسي للأرض على حجب الإشعاعات الضارة،
مما يسمح بوجود الحياة على سطح الأرض. تحجب طبقة الأوزون الأشعة فوق البنفسجية، ويعمل المجال المغناطيسي للأرض على
إزاحة وإبعاد الجسيمات الأولية المشحونة القادمة من الشمس بسرعات عظيمة ويبعدها في الفضاء الخارجي بعيدا عن الأرض، فلا
تتسبب في الإضرار بالكائنات الحية.

أدت الخصائص الفيزيائية للأرض والمدار الفلكي المناسب التي تدور فيه حول الشمس حيث تمدها بالدفء والطاقة ووجود الماء
إلى نشأة الحياة واستمرار الحياة عليها حتى العصر الحالي، ومن المتوقع أن تستمر الحياة على الأرض لمدة 1.2 مليارات عام
آخر، يقضي بعدها ضوء الشمس المتزايد على الغلاف الحيوي للأرض، حيث يعتقد العلماء بأن الشمس سوف ترتفع درجة حرارتها في
المستقبل وتتمدد وتكبر حتى تصبح عملاقا أحمرا ويصل قطرها إلى كوكب الزهرة أو حتى إلى مدار الأرض، على نحو ما يروه من
تطور للنجوم المشابهة للشمس في الكون عند قرب انتهاء عمر النجم ونفاذ وقوده من الهيدروجين. عندئذ تنهي حرارة الشمس
المرتفعة الحياة على الأرض. هذا إذا لم يحدث لها حدث كوني آخر قبل ذلك - كانفجار نجم قريب في هيئة مستعر أعظم - ينهي
الحياة عليها.

يعيش أكثر من 8 مليار شخص على الأرض، وتعمل موارد الأرض المختلفة على إبقاء جمهرة عالمية ضخمة من البشر، الذين يقتسمون
العالم فيما بينهم ويتوزعون على حوالي 200 دولة مستقلة، وطوّر البشر مجتمعات وحضارات وثقافات متنوعة، ويتفاعلون مع
بعضهم البعض بأساليب متنوعة تشمل التواصل الدبلوماسي السياحة التجارة والقتال العسكري أيضًا. ظهر في الثقافة البشرية
نظريات وتمثيلات مختلفة للأرض، فبعض الحضارات القديمة جسدتها كإلهة، والبعض اعتقدها مسطحة، وقال آخرون أنها مركز الكون،
والمعتقد السائد حاليًا ينص على أن هذا الكوكب هو عبارة عن بيئة متكاملة تتطلب إشراف الإنسان عليها لصيانتها من الأخطار
التي تهددها، والتي من شأنها أن تهدد الإنسان نفسه في نهاية المطاف.
'''},

{'language': 'Japanese',
 'main_script': 'Han',
 'script_variant': 'ja',
 'expected_family_names': ['Noto Sans CJK JP', 'Noto Serif CJK JP'],
 'expected_family_name': 'Noto Sans CJK JP',
 'is_rtl': False,
 'text':
'''
地球とは人類が住んでいる天体、つまり人類の足元にある天体のことである。「地」という字・概念と「球」という字・概念で
それを表現している。英語（Earth）やラテン語 （Tellus, Terra）など他の言語でも多くは「大地」を表す語が当てられてい
る。日本語において、この星を呼ぶ名である「地球」という単語は、中国語由来である。中国語の「地球」は明朝の西学東漸
（中国語版）期に初めて見られ、イタリア人宣教師マテオ・リッチ（1552年 - 1610年）の『坤輿万国全図』がこの単語が使用さ
れた最初期の資料である[7][8]。清朝後期に西洋の近代科学が中国に入ってくると、大地球体説が中国の人々によって次第に受
け入れられるようになり、「地球」（または地毬）という単語が広く使われるようになった[9][10][11]。当時の新聞申報の創刊
号には「地球説」に関する文章が掲載されている[12]。日本では、江戸時代頃にこの漢語が輸入され、1700年代頃の西洋紀聞や
和漢三才図会に、使用例がある。幕末から明治期には、庶民も使うほどまでに定着した[13][14][15]。

地球は太陽系の惑星の一つである[5]。その形は、ほぼ回転楕円体で、赤道の半径は6378kmほどで、極半径は6357km[5]。（より
精度の高い数字については後述の「物理的性質」の項を参照のこと）その運動に着目すると、365日強で太陽の周囲を一周し、24
時間で1回 自転しており[5]、太陽からの平均距離は1億4960万km[1]。

その内部は大まかに地殻、マントル、核の3部分から成っている。地球全体の平均密度は1cm3当たり5.51gである[1]。表面は大気
に覆われている[5]。

放射性元素による隕石の年代測定と[16]、アポロ計画によって持ち帰られた月の岩石分析から[17]、地球は誕生してから約46億
年経過していると推定される[18]。

太陽系の年齢もまた隕石の年代測定に依拠するので、地球は太陽系の誕生とほぼ同時に形成されたとしてよい。10個程度の火星
サイズの原始惑星の衝突合体によって形成されたと考えられている[19]。

太陽系内の惑星としては、太陽から2天文単位内の位置に存在し、岩石質外層と鉄を主成分とする中心核を持つ「地球型惑星」に
分類され[20]、太陽系の地球型惑星の中で大きさ、質量、密度ともに最大のものである。


水平線を超えて海面に隠れる船組成は地表面からの深さによって異なる。地殻に存在する元素は、酸素（質量比49.5%）とケイ素
（同25.8%）が主体で、以下アルミニウム・鉄・カルシウム・ナトリウム・カリウム・マグネシウムなどの金属元素が含まれる。
この元素別質量百分率はクラーク数として纏められている[21]。ほとんどはケイ酸塩など金属酸化物の形で存在する[21]。

対照的に、中心部分は鉄やニッケルが主体である。地表面の71.1%は液体の水（海）で被われており[22]、地表から上空約100km
までの範囲には窒素・酸素を主成分とする大気がある。大気の組成は高度によって変化する。

地球はほぼ球形であるため、海抜0mの地表面に立った人が一度に見渡せる範囲は水平線が生じる半径3km〜5kmの円の内側に限ら
れる。分かりやすい事例として、遠方に向かって航行する船，長い直線形の橋，水面に立つ送電用鉄塔の列は、水平線に近づく
と下方に沈み込み、海面に隠れてしまうことが挙げられる。また、電離層や通信衛星や中継回線を用いない無線通信にも、水平
線までの見通し距離内でしか通信出来ないと言う制約が生じる。さらに、緯度が変わると夜間に見える天体に違いが発生する。
地球が球体である証拠は生身の人間には実感しにくいため、かつては地球平面説が信じられたこともあった。
'''},

{'language': 'Korean',
 'main_script': 'Hangul',
 'script_variant': '',
 'expected_family_names': ['Noto Sans CJK KR', 'Noto Serif CJK KR'],
 'expected_family_name': 'Noto Sans CJK KR',
 'is_rtl': False,
 'text':
'''
지구(地球, 영어: Earth)는 태양으로부터 세 번째 행성이며, 조금 두꺼운 대기층으로 둘러싸여 있고, 지금까지 발견된
지구형 행성 가운데 가장 크다. 지구는 45억 6700만 년 전 형성되었으며, 용암 활동이 활발했던 지구와 행성 테이아의
격렬한 충돌로 생성되었을 달을 위성으로 둔다. 지구의 중력은 우주의 다른 물체, 특히 태양과 지구의 유일한 자연위성인
달과 상호작용한다. 지구와 달 사이의 중력 작용으로 조석 현상이 발생한다.

지구의 역사 지구의 형성 과정 지구는 약 45억년 전에 형성되었으며, 태양계가 형성되던 시점과 때를 같이한다. 원시
태양계 원반의 태양 가까운 부분에서는 갓 방출되기 시작한 태양의 복사에너지에 의해 휘발성 성분이 제거되면서 규소를
주성분으로 하는 암석 종류와 철, 니켈 등의 금속성분이 남게 된다. 이들은 원시 태양 주위를 공전하면서 합쳐서 그 크기를
불리게 되는데, 어느 정도 몸집과 중력을 가진 것들을 미행성이라고 부른다. 미행성들은 보다 작은 소행성이나 성간 물질을
유인하여 성장하였다. 미행성의 크기가 커지면 성장속도는 가속된다. 크기가 작은 소행성들이 충돌하게 되면 충돌의
충격으로 조각들이 흩어지게 되나, 크기가 큰 것들이 충돌하게 되면 중력이 강하기 때문에 탈출하는 조각들을 회수할 수
있기 때문이다. 이때 생긴 미행성들 중에서 현재까지 남아 있는 것은 5개이다.

원시 지구는 바깥부분이 거의 완전히 녹은 상태를 경험하게 되면서 성장한다. 원시 지구의 열원은 크게 3가지로 설명할 수
있으며, 첫 번째는 소행성의 충돌이다. 소행성의 충돌은 운동에너지를 열에너지로 바꾸어 원시 지구를 뜨겁게 가열했다.
다른 하나는 중력 에너지이다. 원시지구가 충돌로 인한 가열 때문에 조금씩 녹기 시작하자 그 때까지 뒤섞여 있던 철과
규소가 중력에 의해서 서로 분리되기 시작한 것이다. 무거운 철이 중력에너지가 낮은 지구 중심으로 쏠려 내려가면서
굉장한 중력에너지를 열에너지의 형태로 방출한다. 세 번째 열원은 원시 태양계에 충만하던 방사성 동위원소의 붕괴열이다.
지구의 바깥부분이 완전히 녹은 상태를 마그마 바다라고 한다. 마그마 바다의 깊이는 수백 km에 달했다고 여겨진다. 중력
분화가 끝나고,낙하할 소행성들도 거의 정리가 되자 지구는 식기 시작한다. 마그마 바다가 식기 시작하면서 최초의 지각이
형성된다.

대기와 바다의 형성 지구 대기의 역사는 암석과 마그마로부터 방출된 기체들이 지구 주위에 중력으로 묶이면서 시작된다.
이렇게 형성된 대기를 원시 대기라고 한다. 원시 대기를 이루는 물질은 지구를 형성한 소행성과 혜성 따위에 포함되어있던
휘발성 물질로부터 비롯되었다. 지구가 식어가면서 마그마 바다가 식어 고체의 바닥이 다시 형성되고, 혜성에 들어있던
미량의 물은 많은 양의 혜성이 떨어지면서 축적되기 시작했고, 마그마가 식어 고체의 바닥이 형성된 후에 원시 대기의
수증기 성분이 응결하여 비가 내리기 시작하였다. 이 비는 원시 바다를 형성하였다. 이때 땅과 대기에 있던 염분들이 비에
의해 바다로 녹아들어 가면서 바다가 짜게 되었고 소금을 얻을 수 있게 되었다.
'''},

{'language': 'Hindi',
 'main_script': 'Devanagari',
 'script_variant': '',
 'expected_family_names': ['Noto Sans Devanagari', 'Noto Serif Devanagari'],
 'expected_family_name': 'Noto Sans Devanagari',
 'is_rtl': False,
 'text':
'''
पृथ्वी (प्रतीक: 🜨) सौर मण्डल में सूर्य से तीसरा ग्रह है और एकमात्र खगोलीय वस्तु है जो जीवन को आश्रय देने के लिए
जाना जाता है। इसकी सतह का 71% भाग जल से तथा 29% भाग भूमि से ढका हुआ है। इसकी सतह विभिन्न प्लेटों से बनी हुए है। इस
पर जल तीनो अवस्थाओं में पाया जाता है। इसके दोनों ध्रुवों पर बर्फ की एक मोटी परत है।

रेडियोमेट्रिक डेटिंग अनुमान और अन्य सबूतों के अनुसार पृथ्वी की उत्पत्ति 4.54 अरब साल पहले हुई थी। पृथ्वी के इतिहास
के पहले अरब वर्षों के भीतर जीवों का विकास महासागरों में हुआ और पृथ्वी के वायुमण्डल और सतह को प्रभावित करना शुरू कर
दिया जिससे अवायुजीवी और बाद में, वायुजीवी जीवों का प्रसार हुआ। कुछ भूगर्भीय साक्ष्य इंगित करते हैं कि जीवन का आरम्भ
4.1 अरब वर्ष पहले हुआ होगा। पृथ्वी पर जीवन के विकास के दौरान जैवविविधता का अत्यन्त विकास हुआ। हजारों प्रजातियाँ
लुप्त होती गयी और हजारों नई प्रजातियाँ उत्पन्न हुई। इसी क्रम में पृथ्वी पर रहने वाली 99% से अधिक प्रजातियाँ विलुप्त
हैं। सूर्य से उत्तम दूरी, जीवन के लिए उपयुक्त जलवायु और तापमान ने जीवों में विविधता को बढ़ाया।

पृथ्वी का वायुमण्डल कई परतों से बना हुआ है। नाइट्रोजन और ऑक्सीजन की मात्रा सबसे अधिक है। वायुमण्डल में ओज़ोन गैस की
एक परत है जो सूर्य से आने वाली हानिकारक पराबैंगनी किरणों को रोकती है। वायुमण्डल के घने होने से इस सूर्य का प्रकाश
कुछ मात्रा में प्रवर्तित हो जाता है जिससे इसका तापमान नियन्त्रित रहता है। अगर कोई उल्का पिण्ड पृथ्वी के वायुमण्डल
में प्रवेश कर जाता है तो वायु के घर्षण के कारण या तो जल कर नष्ट हो जाता है या छोटे टुकड़ों में विभाजित हो जाता है।

पृथ्वी की ऊपरी सतह कठोर है। यह पत्थरों और मृदा से बनी है। पृथ्वी का भूपटल कई कठोर खण्डों या विवर्तनिक प्लेटों में
विभाजित है जो भूगर्भिक इतिहास के दौरान एक स्थान से दूसरे स्थान को विस्थापित हुए हैं। इसकी सतह पर विशाल पर्वत, पठार,
महाद्वीप, द्वीप, नदियाँ, समुद्र आदि प्राकृतिक सरंचनाएँ है। पृथ्वी की आन्तरिक रचना तीन प्रमुख परतों में हुई है
भूपटल, भूप्रावार और क्रोड। इसमें से बाह्य क्रोड तरल अवस्था में है और एक ठोस लोहे और निकल के आतंरिक कोर के साथ
क्रिया करके पृथ्वी मे चुम्बकत्व या चुम्बकीय क्षेत्र को पैदा करता है। पृथ्वी का चुम्बकीय क्षेत्र विभिन्न प्रकार के
आवेशित कणों को प्रवेश से रोकता है।

पृथ्वी सूर्य से लगभग 15 करोड़ किलोमीटर दूर स्थित है। दूरी के आधार पर यह सूर्य से तीसरा ग्रह है। यह सौर मण्डल का
सबसे बड़ा चट्टानी पिण्ड है। पृथ्वी सूर्य का एक चक्कर 365 दिनों में पूरा करती है। यह अपने अक्ष पर लम्बवत 23.5 डिग्री
झुकी हुई है। इसके कारण इस पर विभिन्न प्रकार के मौसम आते हैं। अपने अक्ष पर यह 24 घण्टे में एक चक्कर पूरा करती है
जिससे इस पर दिन और रात होती है। चन्द्रमा के पृथ्वी के निकट होने के कारण यह पृथ्वी पर मौसम के लिए दायी है। इसके
आकर्षण के कारण इस पर ज्वार-भाटे उत्पन्न होता है। चन्द्रमा पृथ्वी का एकमात्र प्राकृतिक उपग्रह है।
'''},


{'language': 'Emoji',
 'main_script': 'Common',
 'script_variant': 'Emoji',
 'expected_family_names': ['Noto Color Emoji', 'Apple Color Emoji'] if platform.system() == "Darwin" else 
                          ['Noto Color Emoji', 'Segoe UI Emoji'] if platform.system() == "Windows" else
                          ['Noto Color Emoji'],
 'expected_family_name': 'Apple Color Emoji' if platform.system() == "Darwin" else
                         'Segoe UI Emoji' if platform.system() == "Windows" else
                         'Noto Color Emoji',
 'is_rtl': False,
 'text': # Text below is just some sample emoji
'''
😀😃😄😁🐶🐱🐭🐹🍏🍎🍐🍊⚽️🏀🏈⚾️🚗🚕🚙🚌⌚️📱📲💻🩷❤️🧡💛🏳️🏴🏴‍☠️🏁
'''},
]




'''
Utility script for downloading and generating local copies of Unicode and Noto data for use by fontfinder.
This script is only used during development, and not by users of fontfinder.
'''
import json
from pathlib import Path
import tempfile

import requests

import fontfinder
from fontfinder import noto


def download_noto_ref_data():
    response = requests.get(noto.NOTO_MAIN_JSON_URL)
    with open(noto._NOTO_MAIN_JSON_REF_PATH, "w", encoding="utf-8") as file:
        file.write(response.text)

def download_script_metadata_ref():
    response = requests.get(fontfinder._SCRIPT_METADATA_URL)
    with open(fontfinder._SCRIPT_METADATA_PATH, "w", encoding="utf-8") as file:
        file.write(response.text)

def generate_small_unihan():
    '''Creates the subset of the Unicode Unihan database needed by `fontfinder`.'''
    import unihan_etl.core

    with tempfile.TemporaryDirectory() as full_unihan_dir:
        full_unihan_path = Path(full_unihan_dir, "full_unihan.json").resolve()

        with tempfile.TemporaryDirectory() as work_dir:
            packager_options = {
                "destination": str(full_unihan_path),
                "work_dir": work_dir,
                "format": "json"
            }
            packager = unihan_etl.core.Packager(packager_options)
            packager.download()
            packager.export()

        with open(full_unihan_path) as full_unihan_file:
            with open(fontfinder._SMALL_UNIHAN_PATH, "w", encoding="utf-8") as small_unihan_file:
                full_records = json.load(full_unihan_file)
                selected_keys = ['kTraditionalVariant', 'kSimplifiedVariant']
                small_records = {}
                for full_record in full_records:
                    small_entry = {key: value for key, value in full_record.items() if key in selected_keys}
                    if len(small_entry) > 0:
                        small_records[full_record['char']] = small_entry
                json.dump(small_records, small_unihan_file)
                print(f"Save small Unihan data to {fontfinder._SMALL_UNIHAN_PATH}")
        

if __name__ == '__main__':
    download_noto_ref_data()
    download_script_metadata_ref()
    generate_small_unihan()

